# Copyright contributors to the TSFM project
#
"""Helper class for anomaly detection support"""

from collections import OrderedDict
from typing import List, Optional

import numpy as np
import torch
from sklearn.preprocessing import MinMaxScaler as MinMaxScaler_
from torch import nn as nn
from transformers.utils.generic import ModelOutput

from tsfm_public.models.tinytimemixer.modeling_tinytimemixer import TinyTimeMixerForPrediction
from tsfm_public.toolkit.ad_helpers import AnomalyScoreMethods, ScoreListType, TSADHelperUtility
from tsfm_public.toolkit.conformal import PostHocProbabilisticProcessor


class TinyTimeMixerADUtility(TSADHelperUtility):
    """Implements TSAD Helper Utility for TSPulse model"""

    def __init__(
        self,
        model: TinyTimeMixerForPrediction,
        mode: List[str],
        score_exponent: float = 1.0,
        least_significant_scale: float = 1e-2,
        least_significant_score: float = 0.2,
        probabilistic_processor: Optional[PostHocProbabilisticProcessor] = None,
        **kwargs,
    ):
        """Initializer

        Args:
            model (TinyTimeMixerForPrediction): model instance
            mode (list): list of mode string specifies scoring logic
            score_exponent(float, optional): parameter to sharpen the anomaly score. Defaults to 1.
            least_significant_scale (float, optional): allowed model deviation from the data in the scale of data variance. Defaults to 1e-2.
            least_significant_score (float, optional): minimum anomaly score for significant detection. Defaults to 0.2.

        Raises:
            ValueError: unsupported scoring mode
        """
        if mode is None:
            mode = [AnomalyScoreMethods.PREDICTIVE.value]
        super(TinyTimeMixerADUtility, self).__init__()
        if not self.is_valid_mode(mode):
            raise ValueError(f"Error: unsupported inference method {mode}!")
        self._model = model
        self._mode = mode
        self._score_exponent = score_exponent
        self._least_significant_scale = least_significant_scale
        self._least_significant_score = least_significant_score
        self._probabilistic_processor = probabilistic_processor

    def is_valid_mode(
        self,
        mode_str: List[str],
    ) -> bool:
        """Validates compatibility of the specified mode string."""
        supported_modes = [
            AnomalyScoreMethods.PREDICTIVE.value,
            AnomalyScoreMethods.MEAN_DEVIATION.value,
            AnomalyScoreMethods.PROBABILISTIC.value,
        ]
        for mode_type in supported_modes:
            if mode_type in mode_str:
                return True
        return False

    def compute_score(
        self,
        payload: dict,
        expand_score: bool = False,
        **kwargs,
    ) -> ModelOutput:
        """Produces required model output for anomaly scoring

        Args:
            payload (dict): data batch
            expand_score (bool, optional): compute score for each stream for multivariate data. Defaults to False.

        Returns:
            ModelOutput: model output
        """
        mode = kwargs.get("mode", self._mode)
        use_forecast = AnomalyScoreMethods.PREDICTIVE.value in mode
        use_meandev = AnomalyScoreMethods.MEAN_DEVIATION.value in mode
        use_probabilistic = AnomalyScoreMethods.PROBABILISTIC.value in mode
        anomaly_criterion = nn.MSELoss(reduction="none")

        model_forward_output = {}
        model_forward_output = self._model(**payload)

        scores = OrderedDict()
        batch_future_values = payload["future_values"]
        future_predictions = model_forward_output["prediction_outputs"]
        if use_forecast:
            # forecast output
            reduction_axis = [1] if expand_score else [1, 2]
            pointwise_score = anomaly_criterion(batch_future_values[:, 0, :], future_predictions[:, 0, :]).unsqueeze(1)
            scores[AnomalyScoreMethods.PREDICTIVE.value] = torch.mean(pointwise_score, dim=reduction_axis)
        if use_meandev:
            deviation = batch_future_values - future_predictions
            # if we are not expanding, we take mean across variate dimension
            # time axis is preserved for later adjustment
            scores[AnomalyScoreMethods.MEAN_DEVIATION.value] = (
                deviation if expand_score else torch.mean(deviation, dim=[2])
            )
        if use_probabilistic:
            outlier_score = self._probabilistic_processor.outlier_score(
                y_gt=batch_future_values.detach().cpu().numpy(),
                y_pred=future_predictions.detach().cpu().numpy(),
                outlier_label=False,
                aggregation=None,  # aggregation is done outside this method
            )
            print(f"Outlier score shape {outlier_score.shape}")
            scores[AnomalyScoreMethods.PROBABILISTIC.value] = (
                outlier_score if expand_score else np.mean(outlier_score, axis=2)
            )

        return ModelOutput(scores)

    def adjust_boundary(
        self,
        key: str,
        x: ScoreListType,
        reference: Optional[np.ndarray] = None,
        **kwargs,
    ) -> np.ndarray:
        """Combines model outputs with boundary adjustment

        Args:
            key (str): key associated with model output
            x (ScoreListType): model outputs across all batches combined
            reference (Optional[np.ndarray], optional): reference data for score scale adjustment. Defaults to None.

        Returns:
            np.ndarray: combined score
        """
        if isinstance(x, (list, tuple)):
            if (len(x) > 0) and isinstance(x[0], torch.Tensor):
                x = torch.cat(x, axis=0).detach().cpu().numpy()
            else:
                x = np.concatenate(x, axis=0).astype(float)
        elif isinstance(x, torch.Tensor):
            x = x.detach().cpu().numpy()

        score_exponent = self._score_exponent
        start_pad_len = self._model.config.context_length
        end_pad_len = (
            0 if key == AnomalyScoreMethods.MEAN_DEVIATION.value else self._model.config.prediction_length - 1
        )

        if key == AnomalyScoreMethods.PROBABILISTIC.value:
            if len(x.shape) == 2:
                x = np.expand_dims(x, -1)
            x = self._probabilistic_processor.forecast_horizon_aggregation(x)
            score = np.array([x[0]] * start_pad_len + list(x) + [x[-1]] * end_pad_len)  # (?)
            return score

        if key == AnomalyScoreMethods.MEAN_DEVIATION.value:
            n_batches = x.shape[0]
            n_obs = x.shape[1]
            total_length = n_batches + n_obs - 1
            data_dim = (total_length, x.shape[2]) if x.ndim == 3 else total_length
            counters = np.zeros(data_dim)
            predictions = np.zeros(data_dim)
            for i in range(n_batches):
                predictions[i : (i + n_obs)] += x[i]
                counters[i : (i + n_obs)] += 1
            x = (predictions / np.maximum(counters, 1)) ** 2

        score = np.array([x[0]] * start_pad_len + list(x) + [x[-1]] * end_pad_len)
        if score.ndim == 1:
            score = score.reshape(-1, 1)

        min_score = 0.0
        if reference is not None:
            reference_data = np.asarray(reference)
            min_score = self._least_significant_scale * np.nanstd(reference_data, axis=0, keepdims=True) ** 2
            if min_score.shape[-1] != score.shape[-1]:
                min_score = np.nanmax(min_score, axis=-1)
            if key == AnomalyScoreMethods.PREDICTIVE.value:
                min_score = min_score / np.sqrt(2)
        score_ = score.copy()
        score_[np.where(score > min_score)] *= 1 / self._least_significant_score
        scale = 1 if np.any(score > min_score) else self._least_significant_scale
        score = MinMaxScaler_().fit_transform(score_**score_exponent) * scale
        return score



# Copyright contributors to the TSFM project
#
"""Tools for building TTM Predictor that works with GluonTS datasets"""

import copy
import math
import os
import tempfile
from typing import List

import numpy as np
import pandas as pd
import torch
from gluonts.dataset.split import InputDataset, LabelDataset, TrainingDataset
from gluonts.itertools import batcher
from gluonts.model.forecast import QuantileForecast
from scipy.stats import norm
from torch.optim import AdamW
from torch.optim.lr_scheduler import OneCycleLR
from torch.utils.data import ConcatDataset, DataLoader, Subset
from tqdm.auto import tqdm
from transformers import EarlyStoppingCallback, Trainer, TrainingArguments
from transformers.integrations import INTEGRATION_TO_CALLBACK
from transformers.utils import logging

from tsfm_public import (
    TrackingCallback,
    count_parameters,
)
from tsfm_public.toolkit.get_model import TTM_LOW_RESOLUTION_MODELS_MAX_CONTEXT, get_model
from tsfm_public.toolkit.lr_finder import optimal_lr_finder

from .gluonts_data_wrapper import (
    RESOLUTION_MAP,
    TTM_MAX_FORECAST_HORIZON,
    ForecastDataset,
    StandardScalingGluonTSDataset,
    TorchDatasetFromGluonTSTrainingDataset,
    get_freq_mapping,
    impute_series,
)
from .utils import CustomMASETrainer, plot_forecast


logger = logging.get_logger(__name__)

# TTM Constants:
# Fewshot max allowed number of samples
# This is only used when `upper_bound_fewshot_samples=True`.
# For example, if 5% few-shot for a dataset exceeds this number,
# this `FEWSHOT_MAX_NUM_SAMPLES` upper bound will be used.
FEWSHOT_MAX_NUM_SAMPLES = 500_000


class TTMGluonTSPredictor:
    """Wrapper to TTM that can be directly trained, validated, and tested with GluonTS datasets."""

    def __init__(
        self,
        context_length: int,
        prediction_length: int,
        model_path: str = "ibm-granite/granite-timeseries-ttm-r2",
        test_data_label: LabelDataset = None,  # provide this for plotting
        scale: bool = False,
        random_seed: int = 42,
        term: str = None,
        ds_name: str = None,
        out_dir: str = None,
        upper_bound_fewshot_samples: bool = False,
        force_short_context: bool = False,
        min_context_mult: int = 4,
        past_feat_dynamic_real_exist: bool = False,
        num_prediction_channels: int = None,
        use_mask: bool = False,
        freq: str = None,
        use_valid_from_train: bool = True,
        insample_forecast: bool = True,
        insample_use_train: bool = False,
        plot_predictions: bool = False,
        **kwargs,
    ):
        """Initialize a TTMGluonTSPredictor object.

        Args:
            context_length (int): Context length / history length.
            prediction_length (int): Prediction length / forecast horizon.
            model_path (str, optional): Model path / model card link. Defaults to "ibm-granite/granite-timeseries-ttm-r2".
            test_data_label (LabelDataset, optional): Test data label object. Only used for plotting. Defaults to None.
            random_seed (int, optional): Random seed. Defaults to 42.
            term (str, optional): Term (short/medium/long). Defaults to None.
            ds_name (str, optional): Dataset name. Defaults to None.
            out_dir (str, optional): Output directory. Defaults to None.
            upper_bound_fewshot_samples (bool, optional): If True, number of x% fewshot will be upper-bounded to
                FEWSHOT_MAX_NUM_SAMPLES. Defaults to False (recommended).
            force_short_context (bool, optional): If True, it forces to use short context. Defaults to False (recommended).
            min_context_mult (int, optional): If set to n, minimum context of length `prediction_length x n` is needed
                to create a finetune sample. Defaults to 4.
            past_feat_dynamic_real_exist (bool, optional): Use past exogeneous features. Defaults to False.
            num_prediction_channels (int, optional): Number of prediction channels. Defaults to None.
            use_mask (bool, optional): Use observed mask. Defaults to False.
            freq (str, optional): Frequency. Defaults to None.
            use_valid_from_train (bool, optional): Use unused training data for validation (from 100-x% fewshot).
                Defaults to True.
            insample_forecast (bool, optional): Calculate insample statistics for quantile forecasts. Defaults to True.
            insample_use_train (bool, optional): Use training error statistics along with validation error statistics
                for insample statistics calculation. Defaults to False.
            plot_predictions (bool, optional): Plot predictions or not. Defaults to False.
        """
        self.prediction_length = prediction_length
        self.test_data_label = test_data_label
        self.scale = scale
        self.scaler = None
        self.random_seed = random_seed
        self.term = term
        self.ds_name = ds_name
        self.out_dir = out_dir
        self.upper_bound_fewshot_samples = upper_bound_fewshot_samples
        self.force_short_context = force_short_context
        self.min_context_mult = min_context_mult
        self.past_feat_dynamic_real_exist = past_feat_dynamic_real_exist
        self.num_prediction_channels = num_prediction_channels
        self.freq = freq
        self.use_mask = use_mask
        self.use_valid_from_train = use_valid_from_train
        self.insample_forecast = insample_forecast
        self.insample_use_train = insample_use_train
        self.plot_predictions = plot_predictions
        self.quantile_keys = [
            "0.1",
            "0.2",
            "0.3",
            "0.4",
            "0.5",
            "0.6",
            "0.7",
            "0.8",
            "0.9",
            "mean",
        ]

        self.insample_errors = None

        if force_short_context:
            logger.info(f"Forcing short context: H = {prediction_length}, CL={prediction_length * min_context_mult}")

        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        if "dropout" in kwargs and kwargs["dropout"] is None:
            del kwargs["dropout"]
        if "head_dropout" in kwargs and kwargs["head_dropout"] is None:
            del kwargs["head_dropout"]
        if "loss" in kwargs and kwargs["loss"] is not None:
            self.loss = kwargs["loss"]
            if kwargs["loss"] not in ["mse", "mae", "huber", "pinball"]:
                del kwargs["loss"]
        if "prediction_channel_indices" in kwargs:
            self.prediction_channel_indices = kwargs["prediction_channel_indices"]

        # Get model
        self._get_gift_model(model_path, context_length, prediction_length, freq, **kwargs)

    def _get_gift_model(self, model_path: str, context_length: int, prediction_length: int, freq: str, **kwargs):
        """Get suitable TTM model based on context and forecast lengths.

        Args:
            model_path (str): Model card link.
            context_length (int): Context length.
        """
        self.ttm = None

        prefer_l1_loss = False
        prefer_longer_context = True
        freq_prefix_tuning = False
        force_return = "zeropad"
        if self.term == "short" and (
            str(self.freq).startswith("W")
            or str(self.freq).startswith("M")
            or str(self.freq).startswith("Q")
            or str(self.freq).startswith("A")
        ):
            prefer_l1_loss = True
            prefer_longer_context = False
            freq_prefix_tuning = True

        if self.term == "short" and str(self.freq).startswith("D"):
            prefer_l1_loss = True
            freq_prefix_tuning = True
            if context_length < 2 * TTM_LOW_RESOLUTION_MODELS_MAX_CONTEXT:
                prefer_longer_context = False
            else:
                prefer_longer_context = True

        if self.term == "short" and str(self.freq).startswith("A"):
            self.insample_use_train = False
            self.use_valid_from_train = False
            force_return = "random_init_small"

        if prediction_length > TTM_MAX_FORECAST_HORIZON:
            force_return = "rolling"

        self.ttm = get_model(
            model_path=model_path,
            context_length=context_length,
            prediction_length=prediction_length,
            prefer_l1_loss=prefer_l1_loss,
            prefer_longer_context=prefer_longer_context,
            resolution=RESOLUTION_MAP.get(freq, "oov"),
            freq_prefix_tuning=freq_prefix_tuning,
            force_return=force_return,
            **kwargs,
        ).to(self.device)

        self.context_length = self.ttm.config.context_length

        self.enable_prefix_tuning = False
        if hasattr(self.ttm.config, "resolution_prefix_tuning"):
            self.enable_prefix_tuning = self.ttm.config.resolution_prefix_tuning
        logger.info(f"The TTM has Prefix Tuning = {self.enable_prefix_tuning}")

    def _process_time_series(self, dataset: TrainingDataset, truncate: bool = True) -> List:
        """
        Processes a time series by truncating initial NaNs and forward filling intermittent NaNs.
        Returns a new truncated dataset, and does not modify the original one.

        Args:
            dataset (TrainingDataset): Every series of of shape [channels, length].
            truncate (bool, optional): Truncate the dataset if True. Defaults to True.

        Returns:
            List: Processed time series, each of shape [channels, truncated_length].
        """
        truncated_dataset = list(copy.deepcopy(dataset))
        for i, item in enumerate(truncated_dataset):
            data = item["target"]

            if data.ndim == 1:
                data = data.reshape(1, -1)  # [channels, length]

            if self.past_feat_dynamic_real_exist:
                if item["past_feat_dynamic_real"].ndim == 1:
                    item["past_feat_dynamic_real"] = item["past_feat_dynamic_real"].reshape(1, -1)
                data = np.vstack((data, item["past_feat_dynamic_real"]))

            truncated_dataset[i]["target"] = data

            if not truncate:
                continue

            # Step 1: Determine the longest stretch of initial NaNs across all channels
            valid_mask = ~np.isnan(data)  # Mask of valid (non-NaN) values

            if valid_mask.all():
                continue  # Continue if no NaN

            first_valid = np.argmax(valid_mask.any(axis=0))  # First col with any valid value across channels
            data = data[:, first_valid:]  # Truncate cols before the first valid col

            # Step 2: Perform forward fill for NaNs
            df = pd.DataFrame(data.T, columns=range(data.shape[0]))
            df = df.ffill(axis=0)

            data = df.values.T
            if data.shape[0] == 1:  # [1, truncated_length]
                data = data.reshape(-1)  # [truncated_length]

            truncated_dataset[i]["target"] = data

        return truncated_dataset

    def compute_quantile_forecasts(self, loader, quantiles):
        all_quantile_forecasts = []

        for batch in tqdm(loader, desc="Processing Batches"):
            forecast_samples, insample_errors, point_forecasts = batch

            insample_errors[insample_errors == 0] = 1e-5  # To prevent division by zero

            # Expand scales for quantiles
            batch_size, seq_len, no_channels = forecast_samples.shape
            num_quantiles = len(quantiles)

            scales = np.expand_dims(insample_errors, axis=1)  # Shape: (batch_size, 1, H, C)
            scales = np.tile(scales, (1, num_quantiles, 1, 1))  # Shape: (batch_size, num_quantiles, H, C)

            # Expand quantiles
            quantiles_expanded = np.reshape(quantiles, (1, num_quantiles, 1, 1))  # Shape: (1, num_quantiles, 1, 1)
            quantiles_expanded = np.tile(
                quantiles_expanded, (batch_size, 1, seq_len, no_channels)
            )  # Shape: (batch_size, num_quantiles, H, C)

            # Expand forecasts
            forecasts_expanded = np.expand_dims(forecast_samples, axis=1)  # Shape: (batch_size, 1, H, C)
            forecasts_expanded = np.tile(
                forecasts_expanded, (1, num_quantiles, 1, 1)
            )  # Shape: (batch_size, num_quantiles, H, C)

            # Compute quantile forecasts
            quantile_forecasts = norm.ppf(
                quantiles_expanded, loc=forecasts_expanded, scale=scales
            )  # Shape: (batch_size, num_quantiles, H, C)

            # Append point forecasts
            final_forecasts = np.concatenate(
                (quantile_forecasts, point_forecasts), axis=1
            )  # Shape: (batch_size, num_quantiles+1, H, C)

            # Collect results for the batch
            all_quantile_forecasts.extend(final_forecasts)

        return all_quantile_forecasts

    def train(
        self,
        train_dataset: TrainingDataset,
        valid_dataset: TrainingDataset,
        batch_size: int = 64,
        optimize_batch_size: bool = True,
        freeze_backbone: bool = False,
        learning_rate: float = None,
        num_epochs: int = 30,
        num_workers: int = 8,
        fewshot_fraction: int = 1.0,
        automate_fewshot_fraction: bool = True,
        automate_fewshot_fraction_threshold: int = 200,
        fewshot_location: str = "rand",  # rand/start/end
        save_model: bool = False,
    ):
        """Finetune the TTM.

        Args:
            train_dataset (TrainingDataset): Training dataset.
            valid_dataset (TrainingDataset): Validation dataset.
            batch_size (int, optional): Batch size. Defaults to 64.
            optimize_batch_size (bool, optional): Optimize batch size based on data size to speed up training.
                Defaults to True.
            freeze_backbone (bool, optional): To freeze TTM backbone. Defaults to False.
            learning_rate (float, optional): Learning rate. Defaults to None.
            num_epochs (int, optional): Number of epochs. Defaults to 30.
            num_workers (int, optional): Number of workers. Defaults to 8.
            fewshot_fraction (int, optional): Few-shot fraction. Defaults to 1.0.
            automate_fewshot_fraction (bool, optional): Automate few-shot fraction. Helps for very small dataset.
                Defaults to True.
            automate_fewshot_fraction_threshold (int, optional): Consider this many samples as very small.
            fewshot_location (str, optional): Fewshot location "rand"/"start"/"end". Default to "rand" (recommended).
            save_model (bool, optional): Save model to `self.out_dir`. Defaults to False.

        Raises:
            ValueError: _description_
        """
        train_dataset_scaled = self._process_time_series(train_dataset)
        valid_dataset_scaled = self._process_time_series(valid_dataset)
        logger.info(f"Number of series: Train = {len(train_dataset_scaled)}, Valid = {len(valid_dataset_scaled)}")

        # Standard scale
        if self.scale:
            self.scaler = StandardScalingGluonTSDataset()
            self.scaler.fit(train_dataset_scaled)
            train_dataset_scaled = self.scaler.transform(train_dataset_scaled)
            valid_dataset_scaled = self.scaler.transform(valid_dataset_scaled)
            logger.info("Global scaling done successfully.")

        temp_dir = tempfile.mkdtemp()

        # create train dataset
        dset_train_actual = TorchDatasetFromGluonTSTrainingDataset(
            train_dataset_scaled,
            self.context_length,
            self.prediction_length,
            force_short_context=self.force_short_context,
            min_context_mult=self.min_context_mult,
            send_freq=self.enable_prefix_tuning,
            freq=self.freq,
        )

        if automate_fewshot_fraction:
            fewshot_data_size = int(fewshot_fraction * len(dset_train_actual))
            if fewshot_data_size < automate_fewshot_fraction_threshold and fewshot_data_size > 10:
                fewshot_fraction = 0.9
                num_epochs = 50  # easy to run large epochs since datasets are small
                logger.info(f"Increasing fewshot fraction to {fewshot_fraction} due to small dataset size.")

        dset_valid_from_train = None
        if fewshot_fraction < 1.0:
            if fewshot_location == "rand":
                dset_train = copy.deepcopy(dset_train_actual)

                # Choose randomly
                rng = np.random.default_rng(seed=self.random_seed)
                if self.upper_bound_fewshot_samples:
                    list_size = min(int(fewshot_fraction * len(dset_train)), FEWSHOT_MAX_NUM_SAMPLES)
                else:
                    list_size = int(fewshot_fraction * len(dset_train))

                lst_fewshot_indx = rng.integers(
                    low=0,
                    high=len(dset_train),
                    size=list_size,
                )

                logger.info(f"Length of orginal train set = {len(dset_train)}")

                dset_train = Subset(dset_train_actual, lst_fewshot_indx)
                logger.info(f"Length of {fewshot_fraction * 100} % train set = {len(dset_train)}")

                if len(dset_train) < 1:
                    raise ValueError(
                        f"Data too small for finetuning in fewshot {fewshot_fraction * 100}%. Resulting in 0 samples."
                    )

                if self.use_valid_from_train:
                    all_indx = list(range(0, len(dset_train_actual)))
                    valid_indx = list(set(all_indx) - set(lst_fewshot_indx))

                    valid_size = min(50_000, len(dset_train_actual) - len(dset_train))

                    valid_indx = np.random.choice(valid_indx, valid_size, replace=False)
                    dset_valid_from_train = Subset(dset_train_actual, valid_indx)
            elif fewshot_location == "end" or fewshot_location == "start":
                # re-define train with fewshot calculated
                dset_train = TorchDatasetFromGluonTSTrainingDataset(
                    train_dataset_scaled,
                    self.context_length,
                    self.prediction_length,
                    fewshot_fraction=fewshot_fraction,
                    fewshot_location=fewshot_location,
                    force_short_context=self.force_short_context,
                    min_context_mult=self.min_context_mult,
                    send_freq=self.enable_prefix_tuning,
                    freq=self.freq,
                )
            else:
                raise ValueError("Wrong fewshot_location.")
        else:
            logger.info("Using 100% train data to finetune the model.")
            # create train dataset
            dset_train = TorchDatasetFromGluonTSTrainingDataset(
                train_dataset_scaled,
                self.context_length,
                self.prediction_length,
                force_short_context=self.force_short_context,
                min_context_mult=self.min_context_mult,
                send_freq=self.enable_prefix_tuning,
                freq=self.freq,
            )

        dset_valid = TorchDatasetFromGluonTSTrainingDataset(
            valid_dataset_scaled,
            self.context_length,
            self.prediction_length,
            last_window_only=True,
            gen_more_samples_for_short_series=False,
            force_short_context=self.force_short_context,
            min_context_mult=self.min_context_mult,
            send_freq=self.enable_prefix_tuning,
            freq=self.freq,
        )
        dset_valid_actual = copy.deepcopy(dset_valid)

        if dset_valid_from_train is not None:
            # Use valid for training since it is more recent data,
            # and use random valid chosen earlier
            dset_train = ConcatDataset((dset_train, dset_valid))
            dset_valid = dset_valid_from_train

        self.train_num_samples = len(dset_train)
        self.valid_num_samples = len(dset_valid)
        logger.info(f"Number of train samples = {self.train_num_samples}, valid samples = {self.valid_num_samples}")

        if freeze_backbone:
            logger.info(
                f"Number of params before freezing backbone = {count_parameters(self.ttm)}",
            )

            # Freeze the backbone of the model
            for param in self.ttm.backbone.parameters():
                param.requires_grad = False

            # Count params
            logger.info(
                f"Number of params after freezing the backbone = {count_parameters(self.ttm)}",
            )

        if optimize_batch_size:
            # Set custom batch size to speed up the training
            # for large data and for better convergence for small data
            if self.ttm.config.num_input_channels < 10:
                batch_size = 64
            else:
                batch_size = 16

            if len(dset_train) <= 1_000:
                batch_size = 8
            elif len(dset_train) > 100_000:
                batch_size = 512

            logger.info(
                f"Using a batch size of {batch_size}, based on number of training samples = {len(dset_train)} and number of channels = {self.ttm.config.num_input_channels}."
            )

        # Find optimal learning rate
        # Use with caution: Set it manually if the suggested learning rate is not suitable
        if learning_rate is None:
            learning_rate, self.ttm = optimal_lr_finder(
                self.ttm,
                dset_train,
                batch_size=batch_size,
                enable_prefix_tuning=self.enable_prefix_tuning,
            )
            logger.info(f"OPTIMAL SUGGESTED LEARNING RATE = {learning_rate}")

        logger.info(f"Using learning rate = {learning_rate}")

        finetune_forecast_args = TrainingArguments(
            output_dir=os.path.join(temp_dir, "output"),
            overwrite_output_dir=True,
            learning_rate=learning_rate,
            num_train_epochs=num_epochs,
            do_eval=True,
            eval_strategy="epoch",
            per_device_train_batch_size=batch_size,
            per_device_eval_batch_size=batch_size,
            dataloader_num_workers=num_workers,
            report_to="none",
            save_strategy="epoch",
            logging_strategy="epoch",
            save_total_limit=1,
            logging_dir=os.path.join(temp_dir, "logs"),  # Make sure to specify a logging directory
            load_best_model_at_end=True,  # Load the best model when training ends
            metric_for_best_model="eval_loss",  # Metric to monitor for early stopping
            greater_is_better=False,  # For loss
            seed=self.random_seed,
            data_seed=self.random_seed,
            eval_accumulation_steps=10,
        )

        # Create the early stopping callback
        early_stopping_callback = EarlyStoppingCallback(
            early_stopping_patience=5,  # Number of epochs with no improvement after which to stop
            early_stopping_threshold=1e-5,  # Minimum improvement required to consider as improvement
        )
        tracking_callback = TrackingCallback()

        # Optimizer and scheduler
        optimizer = AdamW(self.ttm.parameters(), lr=learning_rate)
        scheduler = OneCycleLR(
            optimizer,
            learning_rate,
            epochs=num_epochs,
            steps_per_epoch=math.ceil(len(dset_train) / (batch_size)),
        )

        if self.loss == "mase":
            underlying_dataset = train_dataset.dataset.iterable
            first_entry = next(iter(underlying_dataset))
            freq_ = first_entry["freq"]
            hf_trainer = CustomMASETrainer(
                model=self.ttm,
                args=finetune_forecast_args,
                train_dataset=dset_train,
                eval_dataset=dset_valid,
                callbacks=[early_stopping_callback, tracking_callback],
                optimizers=(optimizer, scheduler),
                freq=freq_,
            )
        else:
            hf_trainer = Trainer(
                model=self.ttm,
                args=finetune_forecast_args,
                train_dataset=dset_train,
                eval_dataset=dset_valid,
                callbacks=[early_stopping_callback, tracking_callback],
                optimizers=(optimizer, scheduler),
            )
        hf_trainer.remove_callback(INTEGRATION_TO_CALLBACK["codecarbon"])

        # Force zeroshot if very few finetune samples
        if self.train_num_samples > 10:
            # Finetune
            hf_trainer.train()

        # Save model
        if save_model:
            if self.out_dir is not None:
                hf_trainer.save_model(os.path.join(self.out_dir, "ttm_model"))
            else:
                raise ValueError("`out_dir` should not be `None` when `save_model=True`.")
        if self.insample_forecast:
            self.insample_errors = self.get_insample_stats(
                hf_trainer=hf_trainer,
                dset_valid=dset_valid_actual,
                dset_train=dset_train_actual,
                global_scaler=self.scaler,
                use_train=self.insample_use_train,
            )

        if self.train_num_samples <= 10:
            raise Exception("Forcing zeroshot since number of finetune samples is very low.")

    def get_insample_stats(
        self,
        hf_trainer,
        dset_valid,
        dset_train,
        global_scaler,
        batch_size=128,
        use_train=False,
    ):
        if len(dset_valid) == 1:
            dataset = ConcatDataset((dset_train, dset_valid))
        elif use_train:
            max_subset_size = 50_000
            if len(dset_train) > max_subset_size:
                rng = np.random.default_rng(seed=self.random_seed)
                lst_fewshot_indx = rng.integers(
                    low=0,
                    high=len(dset_train),
                    size=max_subset_size,
                )
                dset_train_subset = Subset(dset_train, lst_fewshot_indx)
                dataset = ConcatDataset((dset_train_subset, dset_valid))
            else:
                dataset = ConcatDataset((dset_train, dset_valid))
        else:
            dataset = dset_valid

        # Get ground truth stacked
        dl_valid = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        y_true = []
        series_ids = []
        for batch in dl_valid:
            y_true.append(batch["future_values"])
            series_ids.extend(batch["item_id"])
        y_true = torch.cat(y_true).detach().cpu().numpy()
        y_true_unscaled = global_scaler.inverse_transform(y_true, series_ids)

        # Get validation predictions
        valid_preds_out = hf_trainer.predict(dataset)
        y_pred = valid_preds_out.predictions[0]
        y_pred_unscaled = global_scaler.inverse_transform(y_pred, series_ids)

        # Create a pands dataframe
        # Flatten (H, C) into 2D arrays
        # L = y_pred_unscaled.shape[0]
        flattened_predictions = list(y_pred_unscaled)
        flattened_ground_truth = list(y_true_unscaled)

        df = pd.DataFrame(
            {
                "item_id": series_ids,
                "y_true": flattened_ground_truth,
                "y_pred": flattened_predictions,
            }
        )

        df["errors"] = (df["y_true"] - df["y_pred"]).abs()  # absolute error pointwise
        errors = df.groupby(by="item_id")["errors"].mean()  # mean over all samples from a particular series

        logger.info("Successfully, calculated the in-sample statistics.")

        return errors

    def validate(
        self,
        valid_dataset: TrainingDataset,
        batch_size: int = 64,
    ):
        """(Optionally) Validate.

        Args:
            valid_dataset (TrainingDataset): Validation dataset.
            batch_size (int, optional): Batch size. Defaults to 64.

        Returns:
            flat: Validation loss.
        """
        valid_dataset_scaled = self._process_time_series(valid_dataset)
        if self.scale:
            if self.scaler is None:
                self.scaler = StandardScalingGluonTSDataset()
                self.scaler.fit(valid_dataset_scaled)

            valid_dataset_scaled = self.scaler.transform(valid_dataset_scaled)
        else:
            valid_dataset_scaled = valid_dataset

        temp_dir = tempfile.mkdtemp()
        dset_valid = TorchDatasetFromGluonTSTrainingDataset(
            valid_dataset_scaled,
            self.context_length,
            self.prediction_length,
            last_window_only=True,
            force_short_context=self.force_short_context,
            min_context_mult=self.min_context_mult,
            send_freq=self.enable_prefix_tuning,
            freq=self.freq,
        )

        # hf_trainer
        hf_trainer = Trainer(
            model=self.ttm,
            args=TrainingArguments(
                output_dir=temp_dir,
                per_device_eval_batch_size=batch_size,
                report_to="none",
                eval_accumulation_steps=10,
                seed=self.random_seed,
                data_seed=self.random_seed,
            ),
        )

        # evaluate = zero-shot performance
        print("+" * 20, "Zero-shot Test Loss", "+" * 20)
        zeroshot_output = hf_trainer.predict(dset_valid)
        print(zeroshot_output)
        return zeroshot_output["eval_loss"]

    def predict(
        self,
        test_data_input: InputDataset,
        batch_size: int = 64,
    ):
        """Predict.

        Args:
            test_data_input (InputDataset): Test input dataset.
            batch_size (int, optional): Batch size. Defaults to 64.

        Returns:
            float: Eval loss.
        """
        # We do not truncate the initial NaNs during testing since it sometimes
        # results in extremely short length, and inference fails.
        # Hence, in the current implementation the initial NaNs will be converted to zeros.
        test_data_input_scaled = self._process_time_series(test_data_input, truncate=False)
        # Standard scale
        if self.scale:
            test_data_input_scaled = self.scaler.transform(test_data_input_scaled)

        while True:
            try:
                # Generate forecast samples
                forecast_samples = []
                series_ids = []
                for batch in tqdm(batcher(test_data_input_scaled, batch_size=batch_size)):
                    batch_ttm = {}
                    adjusted_batch_raw = []
                    past_observed_mask = []
                    for idx, entry in enumerate(batch):
                        series_ids.append(entry["item_id"])

                        # univariate array of shape (time,)
                        # multivariate array of shape (var, time)
                        # TTM supports multivariate time series
                        if len(entry["target"].shape) == 1:
                            entry["target"] = entry["target"].reshape(1, -1)

                        if self.force_short_context:
                            entry["target"] = entry["target"][:, -self.min_context_mult * self.prediction_length :]

                        entry_context_length = entry["target"].shape[1]
                        num_channels = entry["target"].shape[0]

                        # Pad
                        if entry_context_length < self.ttm.config.context_length:
                            logger.debug("Using zero filling for padding.")
                            # Zero-padding
                            padding = torch.zeros(
                                (
                                    num_channels,
                                    self.ttm.config.context_length - entry_context_length,
                                )
                            )
                            adjusted_entry = torch.cat(
                                (
                                    padding,
                                    torch.tensor(impute_series(entry["target"])),
                                ),
                                dim=1,
                            )
                            mask = torch.ones(adjusted_entry.shape)
                            mask[:, : padding.shape[1]] = 0

                            # observed_mask[idx, :, :(ttm.config.context_length - entry_context_length)] = 0
                        # Truncate
                        elif entry_context_length > self.ttm.config.context_length:
                            adjusted_entry = torch.tensor(
                                impute_series(entry["target"][:, -self.ttm.config.context_length :])
                            )
                            mask = torch.ones(adjusted_entry.shape)
                        # Take full context
                        else:
                            adjusted_entry = torch.tensor(impute_series(entry["target"]))
                            mask = torch.ones(adjusted_entry.shape)

                        adjusted_batch_raw.append(adjusted_entry)
                        past_observed_mask.append(mask.bool())

                    # For TTM channel dimension comes at the end
                    batch_ttm["past_values"] = torch.stack(adjusted_batch_raw).permute(0, 2, 1).to(self.device)
                    if self.use_mask:
                        batch_ttm["past_observed_mask"] = (
                            torch.stack(past_observed_mask).permute(0, 2, 1).to(self.device)
                        )
                    if self.ttm.config.resolution_prefix_tuning:
                        freq_map = get_freq_mapping()
                        batch_ttm["freq_token"] = (
                            torch.ones((batch_ttm["past_values"].shape[0])) * freq_map[self.freq]
                        ).to("cuda")

                    if self.prediction_length > TTM_MAX_FORECAST_HORIZON:
                        batch_ttm["return_loss"] = False

                        recursive_steps = int(np.ceil(self.prediction_length / self.ttm.config.prediction_length))
                        predict_outputs = torch.empty(len(batch), 0, num_channels).to(self.device)
                        with torch.no_grad():
                            for i in range(recursive_steps):
                                model_outputs = self.ttm(**batch_ttm)
                                batch_ttm["past_values"] = torch.cat(
                                    [
                                        batch_ttm["past_values"],
                                        model_outputs["prediction_outputs"],
                                    ],
                                    dim=1,
                                )[:, -self.ttm.config.context_length :, :]
                                if self.use_mask:
                                    batch_ttm["past_observed_mask"] = torch.cat(
                                        [
                                            batch_ttm["past_observed_mask"],
                                            torch.ones(model_outputs["prediction_outputs"].shape)
                                            .bool()
                                            .to(self.device),
                                        ],
                                        dim=1,
                                    )[:, -self.ttm.config.context_length :, :]
                                predict_outputs = torch.cat(
                                    [
                                        predict_outputs,
                                        model_outputs["prediction_outputs"][:, : self.ttm.config.prediction_length, :],
                                    ],
                                    dim=1,
                                )
                        predict_outputs = predict_outputs[:, : self.prediction_length, :]
                    else:
                        model_outputs = self.ttm(**batch_ttm)
                        predict_outputs = model_outputs.prediction_outputs

                    # Accumulate all forecasts
                    forecast_samples.append(predict_outputs.detach().cpu().numpy())

                # list to np.ndarray
                forecast_samples = np.concatenate(forecast_samples)

                if self.scale:
                    # inverse scale
                    if self.past_feat_dynamic_real_exist:
                        forecast_samples = self.scaler.inverse_transform(
                            forecast_samples,
                            series_ids,
                            self.prediction_channel_indices,
                        )
                    else:
                        forecast_samples = self.scaler.inverse_transform(forecast_samples, series_ids)

                if self.prediction_length > TTM_MAX_FORECAST_HORIZON:
                    forecast_samples = forecast_samples[:, :, : self.num_prediction_channels]

                if self.insample_forecast:
                    point_forecasts = np.expand_dims(forecast_samples, 1)

                    self.quantiles = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])

                    # Assuming forecasts, scale, and self.quantiles are defined
                    b, seq_len, no_channels = forecast_samples.shape

                    if self.insample_errors is None:
                        dummy_errors_ = []
                        unq_series_ids = list(np.unique(series_ids))
                        for _ in unq_series_ids:
                            dummy_errors_.append(np.ones((seq_len, no_channels)))
                        self.insample_errors = pd.DataFrame(
                            {"item_id": unq_series_ids, "errors": dummy_errors_}
                        ).set_index("item_id")["errors"]
                        logger.warning("`insample_errors` is `None`. Using a dummy error of `np.ones()`")

                    # happens for H > 720
                    if self.insample_errors.iloc[0].shape[0] < self.prediction_length:
                        for i in range(len(self.insample_errors)):
                            self.insample_errors.iloc[i] = np.concatenate(
                                (
                                    self.insample_errors.iloc[i],
                                    self.insample_errors.iloc[i][
                                        -(self.prediction_length - self.insample_errors.iloc[i].shape[0]) :,
                                        :,
                                    ],
                                )
                            )

                    logger.info(f"Making quantile forecasts for quantiles {self.quantiles}")
                    all_quantile_forecasts = []

                    dataset = ForecastDataset(
                        forecast_samples,
                        series_ids,
                        self.insample_errors,
                        point_forecasts,
                        self.quantiles,
                    )
                    dataloader = DataLoader(
                        dataset,
                        batch_size=batch_size,
                        shuffle=False,
                        collate_fn=lambda x: (
                            np.stack([i[0] for i in x]),  # forecast_samples
                            np.stack([i[1] for i in x]),  # insample_errors
                            np.stack([i[2] for i in x]),  # point_forecasts
                        ),
                    )

                    all_quantile_forecasts = self.compute_quantile_forecasts(dataloader, self.quantiles)

                forecast_samples = np.array(all_quantile_forecasts)
                if forecast_samples.shape[-1] == 1:
                    forecast_samples = np.squeeze(forecast_samples, axis=-1)
                break

            except torch.cuda.OutOfMemoryError:
                print(f"OutOfMemoryError at batch_size {batch_size}, reducing to {batch_size // 2}")
                batch_size //= 2

        # Convert forecast samples into gluonts SampleForecast objects
        #   Array of size (num_samples, prediction_length) (1D case) or
        #   (num_samples, prediction_length, target_dim) (multivariate case)
        sample_forecasts = []
        for item, ts in zip(forecast_samples, test_data_input):
            forecast_start_date = ts["start"] + len(ts["target"])
            sample_forecasts.append(
                QuantileForecast(
                    forecast_arrays=item,
                    start_date=forecast_start_date,
                    forecast_keys=self.quantile_keys,
                    item_id=ts["item_id"],
                )
            )

        if self.plot_predictions:
            # plot random samples
            plot_forecast(
                test_data_input,
                self.test_data_label,
                forecast_samples,
                self.prediction_length,
                self.ds_name,
                self.term,
                self.out_dir,
                probabilistic=self.insample_forecast,
                quantile_keys=self.quantile_keys,
            )

        return sample_forecasts



# Copyright contributors to the TSFM project
#
"""Some basic callbacks for training with HF Trainer"""

import time

import numpy as np
from transformers import TrainerCallback
from transformers.trainer_callback import TrainerControl, TrainerState
from transformers.training_args import TrainingArguments


class TrackingCallback(TrainerCallback):
    """Simple tracking callback that tracks per epoch run times and calculates some statistics after training completes.

    Args:
        verbose (bool, optional): If true, prints additional information at the completion of each epoch. Defaults to False.
    """

    def __init__(self, verbose: bool = False):
        self.verbose = verbose

    def on_train_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        self.all_epoch_times = []
        self.train_start_time = time.time()
        return super().on_train_begin(args, state, control, **kwargs)

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        self.train_end_time = time.time()
        self.mean_epoch_time = np.mean(self.all_epoch_times)
        self.total_train_time = self.train_end_time - self.train_start_time
        self.best_eval_metric = state.best_metric
        print(
            f"[{self.__class__.__name__}] Mean Epoch Time = {self.mean_epoch_time} seconds, Total Train Time = {self.total_train_time}"
        )
        return super().on_train_end(args, state, control, **kwargs)

    def on_epoch_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        self.epoch_start_time = time.time()
        return super().on_epoch_begin(args, state, control, **kwargs)

    def on_epoch_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        self.epoch_end_time = time.time()
        self.last_epoch_time = self.epoch_end_time - self.epoch_start_time
        if self.verbose:
            print(f"[{self.__class__.__name__}] Epoch Time = {self.last_epoch_time} seconds")
        self.all_epoch_times.append(self.last_epoch_time)
        return super().on_epoch_end(args, state, control, **kwargs)



"""Base serivce handler"""

import datetime
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd

from .inference_payloads import (
    BaseMetadataInput,
    BaseParameters,
    PredictOutput,
)
from .service_handler import HandlerFunction, ServiceHandler


LOGGER = logging.getLogger(__file__)


class InferenceHandler(ServiceHandler):
    @classmethod
    def load(
        cls, model_id: str, model_path: Union[str, Path]
    ) -> Tuple["InferenceHandler", None] | Tuple[None, Exception]:
        """Load the handler_config -- the tsfm service config for this model, returning the proper
        handler to use the model.

        model_path is expected to point to a folder containing the tsfm_config.json file. This can be a local folder
        or with a model on the HuggingFace Hub.

        Args:
            model_id (str): A string identifier for the model.
            model_path (Union[str, Path]): The full path to the model, can be a local path or a HuggingFace Hub path.
            handler_function (str): The type of handler, currently supported handlers are defined in the HandlerFunction
                enum.

        """

        return super().load(model_id, model_path, handler_function=HandlerFunction.INFERENCE.value)

    def run(
        self,
        data: pd.DataFrame,
        schema: Optional[BaseMetadataInput] = None,
        parameters: Optional[BaseParameters] = None,
        **kwargs,
    ) -> Tuple[PredictOutput, None] | Tuple[None, Exception]:
        """Perform an inference request

        Args:
            data (pd.DataFrame): A pandas dataframe containing historical data.
            schema (Optional[BaseMetadataInput], optional): Service request schema. Defaults to None.
            parameters (Optional[BaseParameters], optional): Service requst parameters. Defaults to None.

        Returns:
            Tuple[PredictOutput, None] | Tuple[None, Exception]: If successful, returns a tuple containing a PredictionOutput
                object as the first element. If unsuccessful, a tuple with an exception as the second element will be returned.
        """

        if not self.prepared:
            return None, RuntimeError("Service wrapper has not yet been prepared; run `handler.prepare()` first.")

        try:
            result = self.implementation.run(data, schema=schema, parameters=parameters, **kwargs)
            encoded_result = encode_dataframe(result)
            counts = self.implementation.calculate_data_point_counts(
                data, output_data=result, schema=schema, parameters=parameters, **kwargs
            )
            return PredictOutput(
                model_id=str(self.implementation.model_id),
                created_at=datetime.datetime.now().isoformat(),
                results=[encoded_result],
                **counts,
            ), None

        except Exception as e:
            return None, e


def encode_dataframe(result: pd.DataFrame, timestamp_column: str = None) -> Dict[str, List[Any]]:
    if timestamp_column and pd.api.types.is_datetime64_any_dtype(result[timestamp_column]):
        result[timestamp_column] = result[timestamp_column].apply(lambda x: x.isoformat())
    return result.to_dict(orient="list")



# Copyright contributors to the TSFM project
#
"""Tsfminference Runtime"""

import copy
import logging
import os
from typing import Any, Dict, List

import pandas as pd
import psutil
from fastapi import APIRouter, HTTPException
from starlette import status

from tsfm_public.toolkit.util import select_by_index
from tsfminference import TSFM_HISTOGRAM

from . import TSFM_ALLOW_LOAD_FROM_HF_HUB, TSFM_MODEL_DIR
from .constants import API_VERSION
from .dataframe_checks import check
from .dirutil import resolve_model_path
from .errors import error_message
from .inference_handler import InferenceHandler
from .inference_payloads import ForecastingInferenceInput, ForecastingMetadataInput, PredictOutput


FORECAST_PROMETHEUS_TIME_SPENT = TSFM_HISTOGRAM("forecast_time_spent", "Wall clock time histogram.")
FORECAST_PROMETHEUS_CPU_USED = TSFM_HISTOGRAM("forecast_cpu_user", "CPU user time histogram.")
FORECAST_PROMETHEUS_MEMORY_USED = TSFM_HISTOGRAM("forecast_memory_used", "memory used histogram.")

LOGGER = logging.getLogger(__file__)


class InferenceRuntime:
    def add_routes(self, app):
        self.router = APIRouter(prefix=f"/{API_VERSION}/inference", tags=["inference"])
        # /forecasting
        self.router.add_api_route(
            "/forecasting",
            self.forecast,
            methods=["POST"],
            response_model=PredictOutput,
        )
        # /modelspec
        self.router.add_api_route("/modelspec", self._modelspec, methods=["GET"])
        app.include_router(self.router)

    def _modelspec(self, model_id: str):
        model_path = resolve_model_path(TSFM_MODEL_DIR, model_id)
        if not model_path:
            raise HTTPException(status_code=404, detail=f"model {model_id} not found.")
        handler, e = InferenceHandler.load(model_id=model_id, model_path=model_path)
        if handler.handler_config:
            answer = {}
            atts = [
                "multivariate_support",
                "missing_value_support",
                "minimum_context_length",
                "maximum_context_length",
                "maximum_prediction_length",
            ]
            for at in atts:
                if hasattr(handler.handler_config, at):
                    answer[at] = getattr(handler.handler_config, at)
            return answer
        else:
            raise HTTPException(status_code=404, detail=str(e))

    def forecast(self, input: ForecastingInferenceInput):
        LOGGER.info("calling forecast_common")
        start = os.times()
        answer, ex = self._forecast_common(input)
        finish = os.times()
        process = psutil.Process()
        FORECAST_PROMETHEUS_TIME_SPENT.observe(finish.elapsed - start.elapsed)
        FORECAST_PROMETHEUS_CPU_USED.observe(finish.user - start.user)
        FORECAST_PROMETHEUS_MEMORY_USED.observe(process.memory_info().rss / 1e9)

        if ex is not None:
            import traceback

            detail = error_message(ex)
            LOGGER.exception(ex)
            traceback.print_exception(ex)
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=detail)

        LOGGER.info("done, returning.")
        return answer

    def _forecast_common(self, input_payload: ForecastingInferenceInput) -> PredictOutput:
        model_path = resolve_model_path(TSFM_MODEL_DIR, input_payload.model_id)

        if not model_path:
            LOGGER.info(f"Could not find model at path: {model_path}")
            if TSFM_ALLOW_LOAD_FROM_HF_HUB:
                model_path = input_payload.model_id
                LOGGER.info(f"Using HuggingFace Hub: {model_path}")
            else:
                return None, RuntimeError(
                    f"Could not load model {input_payload.model_id} from {TSFM_MODEL_DIR}. If trying to load directly from the HuggingFace Hub please ensure that `TSFM_ALLOW_LOAD_FROM_HF_HUB=1`"
                )

        handler, e = InferenceHandler.load(model_id=input_payload.model_id, model_path=model_path)
        if e is not None:
            return None, e

        parameters = input_payload.parameters
        schema = input_payload.schema

        data, ex = decode_data(input_payload.data, schema)
        if ex:
            return None, ValueError("data:" + str(ex))

        future_data, ex = decode_data(input_payload.future_data, schema)
        if ex:
            return None, ValueError("future_data:" + str(ex))

        handler_config = handler.handler_config
        # collect and check underlying time series lengths
        if getattr(handler_config, "minimum_context_length", None) or getattr(
            handler_config, "maximum_context_length", None
        ):
            if schema.id_columns:
                data_lengths = data.groupby(schema.id_columns)[schema.id_columns].apply(len)
                min_len_index = data_lengths.argmin()
                min_data_length = data_lengths.iloc[min_len_index]
                max_data_length = data_lengths.max()
            else:
                min_data_length = max_data_length = len(data)
            LOGGER.info(
                f"Data length recieved {len(data)}, minimum series length: {min_data_length}, maximum series length: {max_data_length}"
            )

        if getattr(handler_config, "minimum_context_length", None):
            if min_data_length < handler_config.minimum_context_length:
                err_str = f"Data for model {input_payload.model_id} should have time series of length that is at least the required model context length."
                if schema.id_columns:
                    err_str += f"Received {min_data_length} time points for id {data_lengths.index[min_len_index]}, but model requires {handler_config.minimum_context_length} time points"
                else:
                    err_str += f"Received {min_data_length} time points, but model requires {handler_config.minimum_context_length} time points"

                return None, ValueError(err_str)

        # truncate data length
        if getattr(handler_config, "maximum_context_length", None):
            if max_data_length > handler_config.maximum_context_length:
                LOGGER.info(f"Truncating series lengths to {handler_config.maximum_context_length}")
                data = select_by_index(
                    data, id_columns=schema.id_columns, start_index=-handler_config.maximum_context_length
                )

        _, e = handler.prepare(data=data, future_data=future_data, schema=schema, parameters=parameters)
        if e is not None:
            return None, e

        LOGGER.info(f"HANDLER: {type(handler)}")
        output, e = handler.run(data=data, future_data=future_data, schema=schema, parameters=parameters)

        if e is not None:
            return None, e

        return output, None


def decode_data(data: Dict[str, List[Any]], schema: ForecastingMetadataInput) -> pd.DataFrame:
    if not data:
        return None, None

    try:
        df = pd.DataFrame.from_dict(data)

        rc, msg = check(df, schema.model_dump())

        if rc != 0:
            return None, ValueError(msg)

        if (ts_col := schema.timestamp_column) and pd.api.types.is_string_dtype(df[ts_col]):
            df[ts_col] = pd.to_datetime(df[ts_col], format="ISO8601")

        sort_columns = copy.copy(schema.id_columns) if schema.id_columns else []

        if ts_col:
            sort_columns.append(ts_col)
        if sort_columns:
            return df.sort_values(sort_columns), None

    except Exception as ex:
        return None, ValueError(str(ex))

    return df, None



# Copyright contributors to the TSFM project
#

"""Tests the time series preprocessor and functions"""

import pandas as pd
import pytest
from transformers import PatchTSTConfig, PatchTSTForPrediction

from tsfm_public import TinyTimeMixerConfig, TinyTimeMixerForPrediction
from tsfm_public.toolkit.time_series_forecasting_pipeline import (
    TimeSeriesForecastingPipeline,
)
from tsfm_public.toolkit.time_series_preprocessor import DEFAULT_FREQUENCY_MAPPING, TimeSeriesPreprocessor
from tsfm_public.toolkit.util import select_by_index


@pytest.fixture(scope="module")
def ttm_dummy_model(conf=None):
    # model_path = "ibm-granite/granite-timeseries-ttm-v1"

    if conf is None:
        conf = TinyTimeMixerConfig()
    model = TinyTimeMixerForPrediction(conf)

    return model


def test_forecasting_pipeline_defaults():
    model = PatchTSTForPrediction(PatchTSTConfig(prediction_length=3, context_length=33))

    tspipe = TimeSeriesForecastingPipeline(model)

    assert tspipe._preprocess_params["prediction_length"] == 3
    assert tspipe._preprocess_params["context_length"] == 33

    tspipe = TimeSeriesForecastingPipeline(model=model, prediction_length=6, context_length=66)

    assert tspipe._preprocess_params["prediction_length"] == 6
    assert tspipe._preprocess_params["context_length"] == 66


def test_forecasting_pipeline_forecasts(patchtst_base_model, etth_data_base):
    timestamp_column = "date"
    id_columns = []
    target_columns = ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"]
    prediction_length = 96

    model = patchtst_base_model
    context_length = model.config.context_length

    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model,
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        freq="1h",
    )

    test_end_index = 12 * 30 * 24 + 8 * 30 * 24
    test_start_index = test_end_index - context_length

    data = etth_data_base.copy()

    test_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=test_start_index,
        end_index=test_end_index,
    )

    delta = test_data[timestamp_column].iloc[-1] - test_data[timestamp_column].iloc[-2]
    df_fut = pd.DataFrame(
        {
            timestamp_column: pd.date_range(
                test_data[timestamp_column].iloc[-1] + delta,
                freq=delta,
                periods=prediction_length,
            )
        }
    )

    # base case
    forecasts = forecast_pipeline(test_data, future_time_series=df_fut)
    assert forecasts.shape == (1, 2 * len(target_columns) + 1)

    # when we provide no data for future time series, we do internal augmentation
    forecasts_no_future = forecast_pipeline(test_data)
    assert forecasts_no_future.shape == (1, 2 * len(target_columns) + 1)

    # check forecasts match
    assert forecasts_no_future.iloc[0]["OT_prediction"] == forecasts.iloc[0]["OT_prediction"]

    # test that forecasts are properly exploded
    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model,
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        freq="1h",
        explode_forecasts=True,
    )

    forecasts_exploded = forecast_pipeline(test_data)
    assert forecasts_exploded.shape == (prediction_length, len(target_columns) + 1)

    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model,
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        freq="1h",
        batch_size=10,
    )

    test_end_index = 12 * 30 * 24 + 8 * 30 * 24
    test_start_index = test_end_index - context_length - 9

    data = etth_data_base.copy()

    test_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=test_start_index,
        end_index=test_end_index,
    )
    forecasts = forecast_pipeline(test_data)
    assert forecast_pipeline._batch_size == 10
    assert forecasts.shape == (10, 2 * len(target_columns) + 1)


def test_forecasting_pipeline_forecasts_with_preprocessor(patchtst_base_model, etth_data_base):
    timestamp_column = "date"
    id_columns = []
    target_columns = ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"]
    prediction_length = 96

    model = patchtst_base_model
    context_length = model.config.context_length

    data = etth_data_base.copy()
    train_end_index = 12 * 30 * 24

    test_end_index = 12 * 30 * 24 + 8 * 30 * 24
    test_start_index = test_end_index - context_length - 4

    train_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=0,
        end_index=train_end_index,
    )
    test_data = select_by_index(
        data,
        id_columns=id_columns,
        start_index=test_start_index,
        end_index=test_end_index,
    )

    tsp = TimeSeriesPreprocessor(
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        context_length=context_length,
        prediction_length=prediction_length,
        freq="1h",
        scaling=True,
    )

    tsp.train(train_data)

    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model,
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        freq="1h",
        feature_extractor=tsp,
        explode_forecasts=False,
        inverse_scale_outputs=True,
    )

    forecasts = forecast_pipeline(test_data)

    assert forecasts.shape == (
        test_end_index - test_start_index - context_length + 1,
        2 * len(target_columns) + 1,
    )

    # if we have inverse scaled mean should be larger
    assert forecasts["HUFL_prediction"].mean().mean() > 10


def test_frequency_token(ttm_dummy_model, etth_data):
    model = ttm_dummy_model
    train_data, test_data, params = etth_data

    timestamp_column = params["timestamp_column"]
    id_columns = params["id_columns"]
    target_columns = params["target_columns"]
    prediction_length = params["prediction_length"]
    context_length = params["context_length"]

    tsp = TimeSeriesPreprocessor(
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        context_length=context_length,
        prediction_length=prediction_length,
        freq="1h",
        scaling=True,
    )

    tsp.train(train_data)

    assert model.config.resolution_prefix_tuning is False

    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model,
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        freq="1h",
        feature_extractor=tsp,
        explode_forecasts=False,
        inverse_scale_outputs=True,
    )
    assert forecast_pipeline._preprocess_params["frequency_token"] is None

    model.config.resolution_prefix_tuning = True
    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model,
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        freq="1h",
        feature_extractor=tsp,
        explode_forecasts=False,
        inverse_scale_outputs=True,
    )
    assert forecast_pipeline._preprocess_params["frequency_token"] == DEFAULT_FREQUENCY_MAPPING["h"]

    with pytest.raises(ValueError):
        forecast_pipeline = TimeSeriesForecastingPipeline(
            model=model,
            timestamp_column=timestamp_column,
            id_columns=id_columns,
            target_columns=target_columns,
            freq="1h",
            explode_forecasts=False,
            inverse_scale_outputs=True,
        )


def test_prediction_filter_length(etth_data):
    pfl = 10
    conf = TinyTimeMixerConfig(prediction_filter_length=pfl)
    model = TinyTimeMixerForPrediction(config=conf)
    train_data, test_data, params = etth_data

    timestamp_column = params["timestamp_column"]
    id_columns = params["id_columns"]
    target_columns = params["target_columns"]
    prediction_length = params["prediction_length"]
    context_length = params["context_length"]

    tsp = TimeSeriesPreprocessor(
        timestamp_column=timestamp_column,
        id_columns=id_columns,
        target_columns=target_columns,
        context_length=context_length,
        prediction_length=prediction_length,
        freq="1h",
        scaling=True,
    )

    tsp.train(train_data)

    assert model.config.prediction_filter_length == pfl

    forecast_pipeline = TimeSeriesForecastingPipeline(
        model=model, feature_extractor=tsp, explode_forecasts=False, inverse_scale_outputs=True, device="cpu"
    )
    forecasts = forecast_pipeline(train_data.iloc[:200])

    assert len(forecasts[f"{target_columns[0]}_prediction"].iloc[0]) == pfl



"""Light weight checks on data frame contents to keep Q/A happy."""

import itertools
from typing import Any, Dict, Tuple

import pandas as pd


SCHEMA_KEYS = [
    "id_columns",
    "timestamp_column",
    "target_columns",
    "observable_columns",
    "control_columns",
    "conditional_columns",
    "static_categorical_columns",
]  # todo can we dynamically generated this from inference_payloads?


def disjoint_sets(data: pd.DataFrame, schema: Dict[str, Any]) -> Tuple[int, str]:
    #
    items = []
    for ent in SCHEMA_KEYS:
        it = schema[ent] if ent in schema else None
        if isinstance(it, str):
            it = [it]
        if it:
            items.append(it)
    all_combinations = []
    for r in range(1, len(items) + 1):
        combinations = itertools.combinations(items, r)
        all_combinations.extend(combinations)

    # make sure that we form empty sets
    for combo in all_combinations:
        if len(combo) < 2:
            continue
        s = set(combo[0])
        for it in combo[1:]:
            if len(set(it).intersection(s)) > 0:
                return 1, f"you may not specify f{it} whilst it also appears in {s}"
            s = s.union(set(it))

    return 0, None


def id_cols_are_int_or_string_types(data: pd.DataFrame, schema: Dict[str, Any]) -> Tuple[int, str]:
    # did user specify id columns
    id_cols = schema["id_columns"] if "id_columns" in schema else None
    if id_cols:
        dtypes = data.dtypes
        for id in id_cols:
            if not (pd.api.types.is_string_dtype(dtypes[id]) or pd.api.types.is_integer_dtype(dtypes[id])):
                return 1, f"data for identifier column {id} must be a string or integer type."
    return 0, None


def columns_referenced_are_there(data: pd.DataFrame, schema: Dict[str, Any]) -> Tuple[int, str]:
    cols_df = list(data.columns)

    # freq
    for s in SCHEMA_KEYS:
        v = schema.get(s, None)
        if v:
            if isinstance(v, list):
                for ch in v:
                    if ch not in cols_df:
                        return 1, f"schema and data mismatch, '{ch}' is not in {','.join(cols_df)}"
            elif isinstance(v, str):
                if v not in cols_df:
                    return 1, f"schema and data mismatch, '{v}' is not in {','.join(cols_df)}"
            else:
                return 1, f"unexpected type {type(v)} given for schema element {s}"
    return 0, None


# append you checks here. these should be as light weight as possible
# avoid copying data, for example.
ALLCHECKS = [columns_referenced_are_there, id_cols_are_int_or_string_types, disjoint_sets]


def check(data: pd.DataFrame, schema: Dict[str, Any]) -> Tuple[int, str]:
    if len(data) < 1:
        return 0, None
    for f in ALLCHECKS:
        rc, msg = f(data, schema)
        if rc != 0:
            return rc, msg
    return 0, None



# Copyright contributors to the TSFM project
#
"""Functions to identify candidate learning rates"""

import inspect
import os
import tempfile
import uuid
from cmath import inf
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from torch import nn
from torch.nn.parallel import DistributedDataParallel
from torch.optim import AdamW, Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader
from transformers import PreTrainedModel
from transformers.data.data_collator import default_data_collator
from transformers.trainer_utils import RemoveColumnsCollator
from transformers.utils import logging


logger = logging.get_logger(__name__)


def join_path_file(file: Union[str, Path], path: Union[str, Path], ext: str = ""):
    "Return `path/file` if file is a string or a `Path`, file otherwise"
    if not isinstance(file, (str, Path)):
        return file
    if not isinstance(path, Path):
        path = Path(path)
    path.mkdir(parents=True, exist_ok=True)
    return path / f"{file}{ext}"


def get_model(model: Union[nn.Module, PreTrainedModel]):
    "Return the model maybe wrapped inside `model`."
    return model.module if isinstance(model, (DistributedDataParallel, nn.DataParallel)) else model


def save_model(
    path,
    model: Union[nn.Module, PreTrainedModel],
    opt: Optional[Optimizer],
    with_opt: bool = True,
    pickle_protocol: int = 2,
):
    "Save `model` to `file` along with `opt` (if available, and if `with_opt`)"
    if opt is None:
        with_opt = False
    state = get_model(model).state_dict()
    if with_opt:
        state = {"model": state, "opt": opt.state_dict()}
    torch.save(state, path, pickle_protocol=pickle_protocol)


def load_model(
    path, model, opt: Optional[Optimizer] = None, with_opt: bool = False, device: str = "cpu", strict: bool = True
) -> nn.Module:
    "load the saved model"
    state = torch.load(path, map_location=device)
    if not opt:
        with_opt = False
    model_state = state["model"] if with_opt else state
    get_model(model).load_state_dict(model_state, strict=strict)
    if with_opt:
        opt.load_state_dict(state["opt"])
    model = model.to(device)


class LinearLR(_LRScheduler):
    """Linearly increases the learning rate between two boundaries over a number of iterations.

    Arguments:
        optimizer (torch.optim.Optimizer): wrapped optimizer.
        end_lr (float): the final learning rate.
        num_iter (int): the number of iterations over which the test occurs.
        last_epoch (int, optional): the index of last epoch. Default: -1.
    """

    def __init__(self, optimizer: Optimizer, end_lr: float, num_iter: int, last_epoch: int = -1):
        self.end_lr = end_lr
        if num_iter <= 1:
            raise ValueError("`num_iter` must be larger than 1")
        self.num_iter = num_iter
        super(LinearLR, self).__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        r = (self.last_epoch + 1) / (self.num_iter - 1)
        return [base_lr + r * (self.end_lr - base_lr) for base_lr in self.base_lrs]


class ExponentialLR(_LRScheduler):
    """Exponentially increases the learning rate between two boundaries over a number of iterations.

    Arguments:
        optimizer (torch.optim.Optimizer): wrapped optimizer.
        end_lr (float): the final learning rate.
        num_iter (int): the number of iterations over which the test occurs.
        last_epoch (int, optional): the index of last epoch. Default: -1.
    """

    def __init__(self, optimizer: Optimizer, end_lr: float, num_iter: int, last_epoch: int = -1):
        self.end_lr = end_lr
        self.last_epoch = last_epoch
        if num_iter <= 1:
            raise ValueError("`num_iter` must be larger than 1")
        self.num_iter = num_iter
        super(ExponentialLR, self).__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        r = (self.last_epoch + 1) / (self.num_iter - 1)
        return [base_lr * (self.end_lr / base_lr) ** r for base_lr in self.base_lrs]


def valley(lrs: List[float], losses: List[float]) -> float:
    "Suggests a learning rate from the longest valley and returns its index"
    n = len(losses)
    max_start, max_end = 0, 0

    # find the longest valley
    lds = [1] * n
    for i in range(1, n):
        for j in range(0, i):
            if (losses[i] < losses[j]) and (lds[i] < lds[j] + 1):
                lds[i] = lds[j] + 1
            if lds[max_end] < lds[i]:
                max_end = i
                max_start = max_end - lds[max_end]

    sections = (max_end - max_start) / 3
    idx = max_start + int(sections) + int(sections / 2)

    return float(lrs[idx])


class LRFinder:
    def __init__(
        self,
        model: Union[nn.Module, PreTrainedModel],
        opt: Optimizer,
        device: str = "cuda",
        start_lr: float = 1e-7,
        end_lr: float = 10,
        num_iter: int = 100,
        step_mode: str = "exp",
        beta: float = 0.98,
        suggestion: str = "valley",
        enable_prefix_tuning: bool = False,
    ):
        self.model = model
        self.opt = opt
        self.device = device
        self.start_lr, self.end_lr = start_lr, end_lr
        self.num_iter = num_iter
        self.step_mode = step_mode
        if beta >= 1:
            raise ValueError("`num_iter` must be smaller than 1")
        else:
            self.beta = beta
        self.suggestion = suggestion
        self.recorder = {}
        self.enable_prefix_tuning = enable_prefix_tuning

    def save(self, fname: Union[str, Path], path: Union[str, Path], **kwargs) -> str:
        """
        Save model and optimizer state (if `with_opt`) to `self.path/file`
        """
        fname = join_path_file(fname, path, ext=".pth")
        save_model(fname, self.model, getattr(self, "opt", None), **kwargs)
        return fname

    def load(self, fname: Union[str, Path], with_opt: bool = False, strict: bool = True, **kwargs):
        """
        load the model
        """
        load_model(fname, self.model, self.opt, with_opt, device=self.device, strict=strict)

    def before_fit(self):
        self.model.to(self.device)
        self.loss_func = nn.MSELoss()

        self.losses, self.lrs = [], []
        self.best_loss, self.aver_loss = inf, 0
        self.train_iter = 0

        # save model to load back after fitting
        uid = uuid.uuid4().hex
        self.temp_path = self.save(f"current_{uid}", tempfile.gettempdir(), with_opt=False)
        # set base_lr for the optimizer
        self.set_lr(self.start_lr)

        # check num_iter
        if not self.num_iter:
            self.num_iter = len(self.dls.train)
        # if self.num_iter > len(self.dls.train): self.num_iter = len(self.dls.train)

        # Initialize the proper learning rate policy
        if self.step_mode.lower() == "exp":
            self.scheduler = ExponentialLR(self.opt, self.end_lr, self.num_iter)
        elif self.step_mode.lower() == "linear":
            self.scheduler = LinearLR(self.opt, self.end_lr, self.num_iter)
        else:
            raise ValueError("Unsupported `step_mode`.")

        self.model.train()

    def train_batch(self, batch: torch.Tensor):
        # forward + get loss + backward + optimize
        self.loss = self.train_step(batch)
        # zero the parameter gradients
        self.opt.zero_grad()
        # gradient
        self.loss.backward()
        # update weights
        self.opt.step()

    def process_data(self, x: torch.Tensor, y: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        x, y = x.to(self.device), y.to(self.device)
        return x, y

    def train_step(self, batch: Union[Dict[str, Any], torch.Tensor]) -> Tuple[torch.Tensor, float]:
        # get the inputs
        if isinstance(batch, dict):
            signature = inspect.signature(self.model.forward)
            signature_args = list(signature.parameters.keys())

            args = {k: batch[k].to(self.device) for k in signature_args if k in batch}
            pred_outputs = self.model(**args)
            loss = pred_outputs.loss

        else:
            self.xb, self.yb = batch[0], batch[1]
            self.xb, self.yb = self.process_data(self.xb, self.yb)
            pred = self.model(self.xb)
            loss = self.loss_func(self.yb, pred)

        return loss

    def after_batch_train(self):
        self.train_iter += 1
        self.scheduler.step()
        self.lrs.append(self.scheduler.get_last_lr()[0])

        # update smooth loss
        self.smoothing(self.beta)
        if self.smoothed_loss < self.best_loss:
            self.best_loss = self.smoothed_loss
        # Stop if the loss is exploding
        if self.smoothed_loss > 4 * self.best_loss:
            raise KeyboardInterrupt  # stop fit method
        if self.train_iter > self.num_iter:
            raise KeyboardInterrupt  # stop fit method

    def smoothing(self, beta: float):
        # Smooth the loss if beta is specified
        self.aver_loss = beta * self.aver_loss + (1 - beta) * self.loss.detach().item()
        self.smoothed_loss = self.aver_loss / (1 - beta**self.train_iter)
        self.losses.append(self.smoothed_loss)

    def fit(self, n_epochs, train_dataloder):
        try:
            for epoch in range(n_epochs):
                for batch in train_dataloder:
                    self.train_batch(batch)
                    self.after_batch_train()
        except KeyboardInterrupt:
            pass

    def after_fit(self):
        # reset the gradients
        self.opt.zero_grad()
        if self.suggestion == "valley":
            self.suggested_lr = valley(self.lrs, self.losses)
        else:
            raise ValueError(f"Unsupported lr suggestion mechanism '{self.suggestion}'.")
        # recorder the suggested learning rate
        self.recorder["suggested_lr"] = self.suggested_lr
        # load back the model at the previous state
        self.load(self.temp_path)
        os.remove(self.temp_path)

    def set_lr(self, lrs: List[float]):
        if not isinstance(lrs, list):
            lrs = [lrs] * len(self.opt.param_groups)
        if len(lrs) != len(self.opt.param_groups):
            raise ValueError(
                "Length of `lrs` is not equal to the number of parameter groups " + "in the given optimizer"
            )
        # update lr
        for param_group, lr in zip(self.opt.param_groups, lrs):
            param_group["lr"] = lr

    def plot_lr_find(self, plot_save_dir: Optional[str] = None):
        # Third Party
        import matplotlib.pyplot as plt

        if plot_save_dir is not None:
            os.makedirs(plot_save_dir, exist_ok=True)
            save_path = os.path.join(plot_save_dir, "lr_finder.png")
        else:
            save_path = "lr_finder.png"

        fig, ax = plt.subplots(1, 1)
        ax.plot(self.lrs, self.losses)
        ax.set_ylabel("Loss")
        ax.set_xlabel("Learning Rate")
        ax.set_xscale("log")
        plt.grid()
        plt.savefig(save_path)
        plt.close()


def optimal_lr_finder(
    model: torch.nn.Module,
    dset_train: torch.utils.data.Dataset,
    optimizer: Optional[Any] = None,
    device: Optional[str] = "cuda",
    start_lr: float = 1e-7,
    end_lr: float = 10.0,
    num_iter: int = 100,
    batch_size: int = 20,
    shuffle: bool = True,
    step_mode: str = "exp",
    suggestion: str = "valley",
    plot_lr_finder: bool = False,
    plot_save_dir: Optional[str] = None,
    enable_prefix_tuning: bool = False,
):
    """Finds the optimal learning rate that can used to train a model.
    This function returns the optimal learning rate and the original model.
    The returned model should be used for subsequent training, since the original models'
    weights will be changed during this procedure. Note that sometimes the return learning
    rate can be very low. In that case, it is suggested to manually set the learning rate
    instead of using this function.
    Example usage.
    ```python
    >>> from tsfm.utils.lr_finder import optimal_lr_finder
    >>> LR, model = optimal_lr_finder(model, dset_train, batch_size=20)
    >>> print("OPTIMAL SUGGESTED LEARNING RATE =", LR)
    ```

    Args:
        model (`nn.Module`): The pytorch model.
        dset_train (`torch.utils.data.Dataset`): The train dataset.
        optimizer (`torch.optim.optimizer.Optimizer`, *optional*, Defaults to None.): Optimizer.
        device (`str`, *optional*, Defaults to "cuda".): Device to use ("cpu"/"cuda").
        start_lr (`float`, *optional*, Defaults to 1e-7.): Start learning rate in search space.
        end_lr (`float`, *optional*, Defaults to 10.): End learning rate in search space.
        num_iter (`int`, *optional*, Defaults to 100.): Number of batch updates.
        batch_size (`int`, *optional*, Defaults to 20.): Batch size for the dataloader.
        shuffle (`bool`, *optional*, Defaults to False.): Whether to shuffle the dataloder.
        step_mode (`str`, *optional*, Defaults to "exp".): Type of learning rate scheduler ("exp"/"linear").
        suggestion (`str`, *optional*, Defaults to "valley".): Learning rate suggestion method (only "valley" is suggested currently).
        plot_lr_finder (`bool`, *optional*, Defaults to False.): Plot the search results.
        plot_save_dir (`str`, *optional*, Defaults to `None`.): Plot the search results.

    Returns:
        `float`: The optimal learning rate.
        `nn.Module`: The original model. This returned model should be used for subsequent training.
    """

    logger.info(
        "LR Finder: Running learning rate (LR) finder algorithm. If the suggested LR is very low, we suggest setting the LR manually."
    )

    # default to cuda, our preference in order is: cuda, mps, cpu
    if device is None:
        device = "cuda"

    # check that selected device is valid
    if device == "cuda" and not torch.cuda.is_available():
        device = "mps"

    if device == "mps" and not torch.backends.mps.is_available():
        device = "cpu"

    # device is valid, now set appropriately
    if device == "cuda":
        device = torch.device(f"cuda:{torch.cuda.current_device()}")
    else:
        device = torch.device(device)

    logger.info(f"LR Finder: Using {device}.")

    # create the right collator in the style of HF
    signature = inspect.signature(model.forward)
    signature_columns = list(signature.parameters.keys())
    data_collator = default_data_collator

    remove_columns_collator = RemoveColumnsCollator(
        data_collator=data_collator,
        signature_columns=signature_columns,
        logger=None,
        description=None,
        model_name=model.__class__.__name__,
    )

    dl_train = DataLoader(
        dataset=dset_train, batch_size=batch_size, shuffle=shuffle, collate_fn=remove_columns_collator
    )

    n_epochs = num_iter // len(dl_train) + 1
    if optimizer is None:
        optimizer = AdamW(model.parameters(), 1e-4)
    lr_finder = LRFinder(
        model,
        optimizer,
        device,
        start_lr,
        end_lr,
        num_iter,
        step_mode,
        suggestion=suggestion,
        enable_prefix_tuning=enable_prefix_tuning,
    )
    lr_finder.before_fit()
    lr_finder.fit(n_epochs=n_epochs, train_dataloder=dl_train)
    lr_finder.after_fit()

    if plot_lr_finder:
        os.makedirs(plot_save_dir, exist_ok=True)
        lr_finder.plot_lr_find(plot_save_dir=plot_save_dir)

    logger.info(f"LR Finder: Suggested learning rate = {lr_finder.suggested_lr}")

    return lr_finder.suggested_lr, lr_finder.model



# Copyright contributors to the TSFM project
#
"""Hugging Face Pipeline for Time Series Imputation"""

from collections import OrderedDict
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import torch
from transformers import PreTrainedModel
from transformers.pipelines.base import (
    GenericTensor,
    build_pipeline_init_args,
)
from transformers.utils import add_end_docstrings, logging

from .dataset import ForecastDFDataset
from .time_series_forecasting_pipeline import TimeSeriesPipeline


logger = logging.get_logger(__name__)

# Eventually we should support all time series models
MODEL_FOR_TIME_SERIES_IMPUTATION_MAPPING_NAMES = OrderedDict(
    [
        # Model for Time Series Imputation
        ("TSPulse", "TSPulseForReconstruction"),
    ]
)


@add_end_docstrings(
    build_pipeline_init_args(has_tokenizer=False, has_feature_extractor=True, has_image_processor=False)
)
class TimeSeriesImputationPipeline(TimeSeriesPipeline):
    """
    Time Series Forecasting using HF time series forecasting models. This pipeline consumes a `pandas.DataFrame`
    containing the time series data and produces a new `pandas.DataFrame` containing the forecasts.

    """

    def __init__(
        self,
        model: Union["PreTrainedModel"],
        *args,
        inverse_scale_outputs: bool = True,
        add_known_ground_truth: bool = True,
        **kwargs,
    ):
        kwargs["inverse_scale_outputs"] = inverse_scale_outputs
        kwargs["add_known_ground_truth"] = add_known_ground_truth

        # autopopulate from feature extractor and model
        if "feature_extractor" in kwargs:
            for p in [
                "id_columns",
                "timestamp_column",
                "target_columns",
                # "observable_columns",
                # "control_columns",
                # "conditional_columns",
                # "categorical_columns",
                # "static_categorical_columns",
            ]:
                if p not in kwargs:
                    kwargs[p] = getattr(kwargs["feature_extractor"], p)

        if "context_length" not in kwargs:
            kwargs["context_length"] = model.config.context_length

        super().__init__(model, *args, **kwargs)

        self.__context_memory = {}
        # these control the shared run_single method
        self._model_output_key = "reconstruction_outputs"
        self._copy_dataset_keys = False

        if self.framework == "tf":
            raise ValueError(f"The {self.__class__} is only available in PyTorch.")

        self.check_model_type(MODEL_FOR_TIME_SERIES_IMPUTATION_MAPPING_NAMES)

    def _sanitize_parameters(
        self,
        **kwargs,
    ):
        """Assigns parameters to the different steps of the process. If context_length and prediction_length
        are not provided they are taken from the model config.

        For expected parameters see the call method below.
        """

        preprocess_kwargs = {}
        postprocess_kwargs = {}

        preprocess_params = [
            "context_length",
            "id_columns",
            "timestamp_column",
            "target_columns",
            # "observable_columns",
            # "control_columns",
            # "conditional_columns",
            # "categorical_columns",
            # "static_categorical_columns",
            "impute_method",
        ]
        postprocess_params = [
            "prediction_length",
            "context_length",
            "id_columns",
            "timestamp_column",
            "target_columns",
            # "observable_columns",
            # "control_columns",
            # "conditional_columns",
            # "categorical_columns",
            # "static_categorical_columns",
            "inverse_scale_outputs",
            "add_known_ground_truth",
        ]

        for c in preprocess_params:
            if c in kwargs:
                preprocess_kwargs[c] = kwargs[c]

        for c in postprocess_params:
            if c in kwargs:
                postprocess_kwargs[c] = kwargs[c]

        # same logic as HF Pipeline
        batch_size = kwargs.get("batch_size", self._batch_size)
        if batch_size is None:
            if self._batch_size is None:
                batch_size = 1
            else:
                batch_size = self._batch_size

        num_workers = kwargs.get("num_workers", self._num_workers)
        if num_workers is None:
            if self._num_workers is None:
                num_workers = 0
            else:
                num_workers = self._num_workers

        forward_kwargs = {"batch_size": batch_size, "num_workers": num_workers}

        return preprocess_kwargs, forward_kwargs, postprocess_kwargs

    def __call__(
        self,
        time_series: Union["pd.DataFrame", str],
        **kwargs,
    ):
        """Main method of the imputation pipeline. Takes the input time series data (in tabular format) and
        produces predictions.

        Args:
            time_series (Union[&quot;pandas.DataFrame&quot;, str]): A pandas dataframe containing the time series on
                which to perform inference.

        Keyword arguments:
            future_time_series (Union[&quot;pandas.DataFrame&quot;, str]): A pandas dataframe containing future values,
                i.e., exogenous or supporting features which are known in advance.

            feature_extractor (TimeSeriesPreprocessor): A time series preprpocessor object that specifies how the time
                series should be prepared. If this is provided, any of the other options below will be automatically
                populated from this instance.

            timestamp_column (str): The name of the column containing the timestamp of the time series.

            id_columns (List[str]): List of column names which identify different time series in a multi-time series input.

            target_columns (List[str]): List of column names which identify the target channels in the input, these are the
                columns that will be forecasted.

            observable_columns (List[str]): List of column names which identify the observable channels in the input.
                Observable channels are channels which we have knowledge about in the past and future. For example, weather
                conditions such as temperature or precipitation may be known or estimated in the future, but cannot be
                changed.

            control_columns (List[str]): List of column names which identify the control channels in the input. Control
                channels are similar to observable channels, except that future values may be controlled. For example, discount
                percentage of a particular product is known and controllable in the future.

            conditional_columns (List[str]): List of column names which identify the conditional channels in the input.
                Conditional channels are channels which we know in the past, but do not know in the future.

            context_length (int): Specifies the length of the context windows extracted from the historical data for feeding into
                the model.

            inverse_scale_outputs (bool): If true and a valid feature extractor is provided, the outputs will be inverse scaled.

            add_known_ground_truth (bool): If True add columns containing the ground truth data (possibly containing missing NaN values) to the imputed columns. Imputed columns will have a
                suffix of "_imputed". These columns have original non-missing values from the ground truth and imputed (reconstructed)
                values from the model at the missing positions in the ground truth. Default True. If False, only columns containing
                imputed values in place of missing NaN values and original non-missing values intact are produced, no suffix is added.

        Return (pandas dataframe):
            A new pandas dataframe containing the imputed series. Each row will contain the id, timestamp, the original
            input feature values and the output forecast for each input column. The output forecast is a list containing
            all the values over the prediction horizon.

        """

        return super().__call__(time_series, **kwargs)

    def preprocess(self, time_series, **kwargs) -> Dict[str, Union[GenericTensor, List[Any]]]:
        """Preprocess step
        Load the data, if not already loaded, and then generate a pytorch dataset.
        """

        timestamp_column = kwargs.get("timestamp_column")

        if isinstance(time_series, str):
            time_series = pd.read_csv(
                time_series,
                parse_dates=[timestamp_column],
            )

        if self.feature_extractor:
            time_series_prep = self.feature_extractor.preprocess(time_series)

        # use forecasting dataset to do the preprocessing
        dataset = ForecastDFDataset(
            time_series_prep,
            prediction_length=0,
            **kwargs,
        )

        # save our series to make life easier
        self.__context_memory["data"] = time_series

        return {"dataset": dataset}

    def _forward(self, model_inputs, **kwargs):
        """Forward step
        Responsible for taking pre-processed dictionary of tensors and passing it to
        the model. Aligns model parameters with the proper input parameters. Only passes
        the needed parameters from the dictionary to the model, but adds them back to the
        ouput for the next step.

        The keys in model_outputs are governed by the underlying model combined with any
        original input keys.
        """

        model_outputs = self.model(**model_inputs)

        return model_outputs

    def postprocess(self, input, **kwargs):
        """Postprocess step
        Takes the dictionary of outputs from the previous step and converts to a more user
        readable pandas format.

        If the explode forecasts option is True, then individual forecasts are expanded as multiple
        rows in the dataframe. This should only be used when producing a single forecast (i.e., unexploded
        result is one row per ID).
        """

        out = self.__context_memory["data"].copy()  # original dataframe
        # input is a list of tensors: bs x context_length x features
        input = torch.cat(input, axis=0).detach().cpu().numpy()

        # stitching logic
        n_batches = input.shape[0]
        n_obs = input.shape[1]
        total_length = n_batches + n_obs - 1
        data_dim = (total_length, input.shape[2]) if input.ndim == 3 else total_length
        counters = np.zeros(data_dim)
        predictions = np.zeros(data_dim)
        for i in range(n_batches):
            predictions[i : (i + n_obs)] += input[i]
            counters[i : (i + n_obs)] += 1
        reconstructed_out = predictions / np.maximum(counters, 1)  # this output is all reconstructions from the model

        reconstructed_df = pd.DataFrame(reconstructed_out, columns=kwargs["target_columns"])

        # inverse scale if we have a feature extractor
        if self.feature_extractor is not None and kwargs["inverse_scale_outputs"]:
            reconstructed_df = self.feature_extractor.inverse_scale_targets(reconstructed_df)

        # need to select original values for non-missing points and use the reconstructed values only for missing points
        reconstructed_df.index = out.index
        imputed_output = out.where(~out.isna(), reconstructed_df)

        dtype_map = {col: out[col].dtype for col in out.columns}
        imputed_output = imputed_output.astype(
            dtype_map
        )  # setting the type of imputed columns to same as the input data columns

        if kwargs["add_known_ground_truth"]:
            imputed_output = imputed_output[kwargs["target_columns"]]
            imputed_output = imputed_output.add_suffix("_imputed")
            out = pd.concat([out, imputed_output], axis=1)
        else:
            out = imputed_output

        self.__context_memory = {}
        return out



# Copyright contributors to the TSFM project
#
"""Utilities for handling datasets"""

import logging
from importlib import resources
from pathlib import Path
from typing import Optional

import pandas as pd
import yaml

from .time_series_preprocessor import TimeSeriesPreprocessor, get_datasets


LOGGER = logging.getLogger(__file__)


def load_dataset(
    dataset_name: str,
    context_length,
    forecast_length,
    fewshot_fraction=1.0,
    fewshot_location="first",
    dataset_root_path: str = "datasets/",
    dataset_path: Optional[str] = None,
    use_frequency_token: bool = False,
    enable_padding: bool = True,
    seed: int = 42,
    **dataset_kwargs,
):
    LOGGER.info(f"Dataset name: {dataset_name}, context length: {context_length}, prediction length {forecast_length}")

    config_path = resources.files("tsfm_public.resources.data_config")
    names_to_config = {p.stem: p for p in config_path.iterdir() if p.suffix == ".yaml"}

    config_path = names_to_config.get(dataset_name, None)

    if config_path is None:
        raise ValueError(
            f"Currently the `load_dataset()` function supports the following datasets: {names_to_config.keys()}\n \
                         For other datasets, please provide the proper configs to the TimeSeriesPreprocessor (TSP) module."
        )

    config = yaml.safe_load(open(config_path, "r"))

    tsp = TimeSeriesPreprocessor(
        id_columns=config["id_columns"],
        timestamp_column=config["timestamp_column"],
        target_columns=config["target_columns"],
        observable_columns=config["observable_columns"],
        control_columns=config["control_columns"],
        conditional_columns=config["conditional_columns"],
        static_categorical_columns=config["static_categorical_columns"],
        scaling=config["scale"]["scaling"],
        scaler_type=config["scale"]["scaler_type"],
        encode_categorical=config["encode_categorical"],
        freq=config["freq"],
        context_length=context_length,
        prediction_length=forecast_length,
    )

    split_config = config["split"]

    # if dataset_path is provided we will ignore the config file
    if dataset_path is None:
        dataset_path = Path(dataset_root_path) / config["data_path"] / config["data_file"]

    data = pd.read_csv(
        dataset_path,
        parse_dates=[config["timestamp_column"]],
    )

    train_dataset, valid_dataset, test_dataset = get_datasets(
        tsp,
        data,
        split_config=split_config,
        fewshot_fraction=fewshot_fraction,
        fewshot_location=fewshot_location,
        use_frequency_token=use_frequency_token,
        enable_padding=enable_padding,
        seed=seed,
        **dataset_kwargs,
    )
    LOGGER.info(f"Data lengths: train = {len(train_dataset)}, val = {len(valid_dataset)}, test = {len(test_dataset)}")

    return train_dataset, valid_dataset, test_dataset



# Copyright contributors to the TSFM project
#
"""Base processor for time series data preparation, and pre/post-processing

Adapted from HF FeatureExtractionMixin, but allows customization of the serialized file name
"""

import json
import logging
import os
import warnings
from typing import Any, Union

from transformers.dynamic_module_utils import custom_object_save
from transformers.feature_extraction_utils import (
    FeatureExtractionMixin,
)
from transformers.utils import (
    cached_file,
    download_url,
    is_offline_mode,
    is_remote_url,
)


LOGGER = logging.getLogger(__file__)


class BaseProcessor(FeatureExtractionMixin):
    PROCESSOR_NAME = "processor_config.json"

    def save_pretrained(self, save_directory: Union[str, os.PathLike], push_to_hub: bool = False, **kwargs):
        """
        Save a feature_extractor object to the directory `save_directory`, so that it can be re-loaded using the
        [`~feature_extraction_utils.FeatureExtractionMixin.from_pretrained`] class method.

        Args:
            save_directory (`str` or `os.PathLike`):
                Directory where the feature extractor JSON file will be saved (will be created if it does not exist).
            push_to_hub (`bool`, *optional*, defaults to `False`):
                Whether or not to push your model to the Hugging Face model hub after saving it. You can specify the
                repository you want to push to with `repo_id` (will default to the name of `save_directory` in your
                namespace).
            kwargs (`Dict[str, Any]`, *optional*):
                Additional key word arguments passed along to the [`~utils.PushToHubMixin.push_to_hub`] method.
        """
        use_auth_token = kwargs.pop("use_auth_token", None)

        if use_auth_token is not None:
            warnings.warn(
                "The `use_auth_token` argument is deprecated and will be removed in v5 of Transformers. Please use `token` instead.",
                FutureWarning,
            )
            if kwargs.get("token", None) is not None:
                raise ValueError(
                    "`token` and `use_auth_token` are both specified. Please set only the argument `token`."
                )
            kwargs["token"] = use_auth_token

        if os.path.isfile(save_directory):
            raise AssertionError(f"Provided path ({save_directory}) should be a directory, not a file")

        os.makedirs(save_directory, exist_ok=True)

        if push_to_hub:
            commit_message = kwargs.pop("commit_message", None)
            repo_id = kwargs.pop("repo_id", save_directory.split(os.path.sep)[-1])
            repo_id = self._create_repo(repo_id, **kwargs)
            files_timestamps = self._get_files_timestamps(save_directory)

        # If we have a custom config, we copy the file defining it in the folder and set the attributes so it can be
        # loaded from the Hub.
        if self._auto_class is not None:
            custom_object_save(self, save_directory, config=self)

        # If we save using the predefined names, we can load using `from_pretrained`
        output_feature_extractor_file = os.path.join(save_directory, self.__class__.PROCESSOR_NAME)

        self.to_json_file(output_feature_extractor_file)
        LOGGER.info(f"Feature extractor saved in {output_feature_extractor_file}")

        if push_to_hub:
            self._upload_modified_files(
                save_directory,
                repo_id,
                files_timestamps,
                commit_message=commit_message,
                token=kwargs.get("token"),
            )

        return [output_feature_extractor_file]

    @classmethod
    def get_feature_extractor_dict(
        cls, pretrained_model_name_or_path: Union[str, os.PathLike], **kwargs
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """
        From a `pretrained_model_name_or_path`, resolve to a dictionary of parameters, to be used for instantiating a
        feature extractor of type [`~feature_extraction_utils.FeatureExtractionMixin`] using `from_dict`.

        Parameters:
            pretrained_model_name_or_path (`str` or `os.PathLike`):
                The identifier of the pre-trained checkpoint from which we want the dictionary of parameters.

        Returns:
            `Tuple[Dict, Dict]`: The dictionary(ies) that will be used to instantiate the feature extractor object.
        """
        cache_dir = kwargs.pop("cache_dir", None)
        force_download = kwargs.pop("force_download", False)
        resume_download = kwargs.pop("resume_download", None)
        proxies = kwargs.pop("proxies", None)
        subfolder = kwargs.pop("subfolder", None)
        token = kwargs.pop("token", None)
        use_auth_token = kwargs.pop("use_auth_token", None)
        local_files_only = kwargs.pop("local_files_only", False)
        revision = kwargs.pop("revision", None)

        if use_auth_token is not None:
            warnings.warn(
                "The `use_auth_token` argument is deprecated and will be removed in v5 of Transformers. Please use `token` instead.",
                FutureWarning,
            )
            if token is not None:
                raise ValueError(
                    "`token` and `use_auth_token` are both specified. Please set only the argument `token`."
                )
            token = use_auth_token

        from_pipeline = kwargs.pop("_from_pipeline", None)
        from_auto_class = kwargs.pop("_from_auto", False)

        user_agent = {"file_type": "feature extractor", "from_auto_class": from_auto_class}
        if from_pipeline is not None:
            user_agent["using_pipeline"] = from_pipeline

        if is_offline_mode() and not local_files_only:
            LOGGER.info("Offline mode: forcing local_files_only=True")
            local_files_only = True

        pretrained_model_name_or_path = str(pretrained_model_name_or_path)
        is_local = os.path.isdir(pretrained_model_name_or_path)
        if os.path.isdir(pretrained_model_name_or_path):
            feature_extractor_file = os.path.join(pretrained_model_name_or_path, cls.PROCESSOR_NAME)
        if os.path.isfile(pretrained_model_name_or_path):
            resolved_feature_extractor_file = pretrained_model_name_or_path
            is_local = True
        elif is_remote_url(pretrained_model_name_or_path):
            feature_extractor_file = pretrained_model_name_or_path
            resolved_feature_extractor_file = download_url(pretrained_model_name_or_path)
        else:
            feature_extractor_file = cls.PROCESSOR_NAME
            try:
                # Load from local folder or from cache or download from model Hub and cache
                resolved_feature_extractor_file = cached_file(
                    pretrained_model_name_or_path,
                    feature_extractor_file,
                    cache_dir=cache_dir,
                    force_download=force_download,
                    proxies=proxies,
                    resume_download=resume_download,
                    local_files_only=local_files_only,
                    subfolder=subfolder,
                    token=token,
                    user_agent=user_agent,
                    revision=revision,
                )
            except OSError:
                # Raise any environment error raise by `cached_file`. It will have a helpful error message adapted to
                # the original exception.
                raise
            except Exception:
                # For any other exception, we throw a generic error.
                raise OSError(
                    f"Can't load feature extractor for '{pretrained_model_name_or_path}'. If you were trying to load"
                    " it from 'https://huggingface.co/models', make sure you don't have a local directory with the"
                    f" same name. Otherwise, make sure '{pretrained_model_name_or_path}' is the correct path to a"
                    f" directory containing a {cls.PROCESSOR_NAME} file"
                )

        try:
            # Load feature_extractor dict
            with open(resolved_feature_extractor_file, encoding="utf-8") as reader:
                text = reader.read()
            feature_extractor_dict = json.loads(text)

        except json.JSONDecodeError:
            raise OSError(
                f"It looks like the config file at '{resolved_feature_extractor_file}' is not a valid JSON file."
            )

        if is_local:
            LOGGER.info(f"loading configuration file {resolved_feature_extractor_file}")
        else:
            LOGGER.info(
                f"loading configuration file {feature_extractor_file} from cache at {resolved_feature_extractor_file}"
            )

        return feature_extractor_dict, kwargs



"""Tuning handler for TSFM models"""

import copy
import logging
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Subset
from transformers import (
    EarlyStoppingCallback,
    Trainer,
    TrainingArguments,
)

from tsfm_public import ForecastDFDataset
from tsfm_public.toolkit.lr_finder import optimal_lr_finder
from tsfm_public.toolkit.time_series_preprocessor import TimeSeriesPreprocessor
from tsfm_public.toolkit.tsfm_config import TSFMConfig
from tsfm_public.toolkit.util import FractionLocation, select_by_fixed_fraction

from .filelogging_tracker import FileLoggingCallback
from .ftpayloads import TuneTypeEnum
from .inference_payloads import ForecastingMetadataInput, ForecastingParameters
from .tsfm_util import load_config, load_model, register_config


LOGGER = logging.getLogger(__file__)


class TSFMForecastingTuningHandler:
    def __init__(
        self,
        model_id: str,
        model_path: Union[str, Path],
        handler_config: TSFMConfig,
    ):
        self.training_label_names = ["future_values"]

        if (
            getattr(handler_config, "model_type", None)
            and getattr(handler_config, "model_config_name", None)
            and getattr(handler_config, "module_path", None)
        ):
            register_config(
                handler_config.model_type,
                handler_config.model_config_name,
                handler_config.module_path,
            )
            LOGGER.info(f"registered {handler_config.model_type}")

        self.model_id = model_id
        self.model_path = model_path
        self.handler_config = handler_config

        # set during prepare
        self.config = None
        self.model = None
        self.preprocessor = None

    def _get_config_kwargs(
        self,
        parameters: Optional[ForecastingParameters] = None,
        preprocessor: Optional[TimeSeriesPreprocessor] = None,
    ) -> Dict[str, Any]:
        """Helper function to return additional configuration arguments that are used during config load.
        Can be overridden in a subclass to allow specialized model functionality.

        Args:
            parameters (Optional[ForecastingParameters], optional): Request parameters. Defaults to None.
            preprocessor (Optional[TimeSeriesPreprocessor], optional): Time seres preprocessor. Defaults to None.

        Returns:
            Dict[str, Any]: Dictionary of additional arguments that are used later as keyword arguments to the config.
        """
        return {"num_input_channels": preprocessor.num_input_channels}

    def prepare(
        self,
        data: pd.DataFrame,
        future_data: Optional[pd.DataFrame] = None,
        schema: Optional[ForecastingMetadataInput] = None,
        parameters: Optional[ForecastingParameters] = None,
        **kwargs,
    ) -> "TSFMForecastingTuningHandler":
        """Implementation of _prepare for HF-like models. We assume the model will make use of the TSFM
        preprocessor and forecasting pipeline. This method:
        1) loades the preprocessor, creating a new one if the model does not already have a preprocessor
        2) updates model configuration arguments by calling _get_config_kwargs
        3) loads the HuggingFace model, passing the updated config object

        Args:
            data (pd.DataFrame): A pandas dataframe containing historical data.
            future_data (Optional[pd.DataFrame], optional): A pandas dataframe containing future data. Defaults to None.
            schema (Optional[ForecastingMetadataInput], optional): Schema information from the original inference
                request. Includes information about columns and their role. Defaults to None.
            parameters (Optional[ForecastingParameters], optional): Parameters from the original inference
                request. Defaults to None.

        Returns:
            ForecastingHuggingFaceHandler: The updated service handler object.
        """

        preprocessor_params = copy.deepcopy(schema.model_dump())
        preprocessor_params["prediction_length"] = parameters.prediction_length

        LOGGER.info(f"Preprocessor params: {preprocessor_params}")

        # load preprocessor
        try:
            preprocessor = TimeSeriesPreprocessor.from_pretrained(self.model_path)
            LOGGER.info("Successfully loaded preprocessor")
        except OSError:
            preprocessor = None
            LOGGER.info("No preprocessor found")
        except Exception as ex:
            raise ex

        if self.handler_config.is_finetuned and preprocessor is None:
            raise ValueError("Model indicates that it is finetuned but no preprocessor was found.")

        if not self.handler_config.is_finetuned and preprocessor is not None:
            raise ValueError("Unexpected: model indicates that it is not finetuned but a preprocessor was found.")

        if preprocessor is None:
            to_check = ["conditional_columns", "control_columns", "observable_columns", "static_categorical_columns"]

            for param in to_check:
                if param in preprocessor_params and preprocessor_params[param]:
                    raise ValueError(
                        f"Unexpected parameter {param} for a zero-shot model, please confirm you have the correct model_id and schema."
                    )

            preprocessor = TimeSeriesPreprocessor(
                **preprocessor_params,
                scaling=False,
                encode_categorical=False,
            )
            # train to estimate freq
            preprocessor.train(data)
            LOGGER.info(f"Data frequency determined: {preprocessor.freq}")
        else:
            # check payload, but only certain parameters
            to_check = [
                "freq",
                "timestamp_column",
                "target_columns",
                "conditional_columns",
                "control_columns",
                "observable_columns",
            ]

            for param in to_check:
                param_val = preprocessor_params[param]
                param_val_saved = getattr(preprocessor, param)
                if param_val != param_val_saved:
                    raise ValueError(
                        f"Attempted to use a fine-tuned model with a different schema, please confirm you have the correct model_id and schema. Error in parameter {param}: received {param_val} but expected {param_val_saved}."
                    )

        model_config_kwargs = self._get_config_kwargs(
            parameters=parameters,
            preprocessor=preprocessor,
        )
        LOGGER.info(f"model_config_kwargs: {model_config_kwargs}")
        model_config = load_config(self.model_path, **model_config_kwargs)

        model = load_model(
            self.model_path,
            config=model_config,
            module_path=self.handler_config.module_path,
        )

        self.config = model_config
        self.model = model
        self.preprocessor = preprocessor

    def train(
        self,
        data: pd.DataFrame,
        schema: ForecastingMetadataInput,
        parameters: ForecastingParameters,
        tuned_model_name: str,
        tmp_dir: Path,
    ) -> str:
        # to do: data selection here
        train_data = data
        validation_data = None
        rng = np.random.default_rng(seed=parameters.random_seed)

        # optional freezing
        if parameters.tune_type == TuneTypeEnum.linear_probe.value:
            # Freeze the backbone of the model
            for param in self.model.backbone.parameters():
                param.requires_grad = False

        # create datasets
        dataset_spec = {
            "id_columns": self.preprocessor.id_columns,
            "timestamp_column": self.preprocessor.timestamp_column,
            "target_columns": self.preprocessor.target_columns,
            "observable_columns": self.preprocessor.observable_columns,
            "control_columns": self.preprocessor.control_columns,
            "conditional_columns": self.preprocessor.conditional_columns,
            "static_categorical_columns": self.preprocessor.static_categorical_columns,
            "prediction_length": self.model.config.prediction_length,
            "context_length": self.model.config.context_length,
        }
        # to do: check prediction length. Three options (1) prediction length is passed, (2) prediction length is in preprocessor, (3) prediction length is in model config
        # we should properly set the prediction/context lengths in the preprocessor
        # we also need to make sure prediction_filter length is set in the model config

        if parameters.fewshot_fraction_location not in [e.value for e in FractionLocation]:
            raise ValueError(
                f"Received unknown location for fewshot_fraction_location {parameters.fewshot_fraction_location}"
            )

        if parameters.fewshot_fraction < 1:
            if parameters.fewshot_fraction_location != FractionLocation.UNIFORM.value:
                train_data = select_by_fixed_fraction(
                    train_data,
                    id_columns=self.preprocessor.id_columns,
                    fraction=parameters.fewshot_fraction,
                    location=parameters.fewshot_fraction_location,  # to do: expose this parameter
                    minimum_size=self.preprocessor.context_length,
                )
            train_dataset = ForecastDFDataset(train_data, **dataset_spec)
            validation_dataset = (
                ForecastDFDataset(validation_data, **dataset_spec) if validation_data else train_dataset
            )

            # uniform case
            if parameters.fewshot_fraction_location == FractionLocation.UNIFORM.value:
                lst = rng.integers(
                    low=0, high=len(train_dataset), size=int(parameters.fewshot_fraction * len(train_dataset))
                )
                train_dataset = Subset(train_dataset, lst.tolist())

        if not parameters.trainer_args.learning_rate:
            learning_rate, self.model = optimal_lr_finder(
                self.model,
                train_dataset,
                batch_size=parameters.trainer_args.per_device_train_batch_size,
                # enable_prefix_tuning=True,
            )
        else:
            learning_rate = parameters.trainer_args.learning_rate

        # Configure trainer
        parameters = parameters
        training_tmp_dir = Path(tmp_dir)
        training_args = TrainingArguments(
            output_dir=training_tmp_dir / "output",
            overwrite_output_dir=True,
            learning_rate=learning_rate,
            num_train_epochs=parameters.trainer_args.num_train_epochs,
            eval_strategy="epoch",
            per_device_train_batch_size=parameters.trainer_args.per_device_train_batch_size,
            per_device_eval_batch_size=parameters.trainer_args.per_device_eval_batch_size,
            dataloader_num_workers=4,  # to do: auto config this parameter
            save_strategy="epoch",
            logging_strategy="epoch",
            save_total_limit=3,
            logging_dir=training_tmp_dir / "logs",  # Make sure to specify a logging directory
            load_best_model_at_end=True,  # Load the best model when training ends
            metric_for_best_model="eval_loss",  # Metric to monitor for early stopping
            greater_is_better=False,  # For loss
            label_names=self.training_label_names,
            use_cpu=not torch.cuda.is_available(),  # is MPS possible now?
        )

        callbacks = [
            FileLoggingCallback(logs_filename=os.environ.get("TSFM_TRAINING_TRACKER_LOGFILE", "training_logs.jsonl"))
        ]
        if parameters.trainer_args.early_stopping and validation_dataset:
            # Create the early stopping callback
            early_stopping_callback = EarlyStoppingCallback(
                early_stopping_patience=parameters.trainer_args.early_stopping_patience,  # Number of epochs with no improvement after which to stop
                early_stopping_threshold=parameters.trainer_args.early_stopping_threshold,  # Minimum improvement required to consider as improvement
            )
            callbacks.append(early_stopping_callback)
        trainer_extra_args = {}
        if validation_dataset:
            trainer_extra_args = {"eval_dataset": validation_dataset}

        # define trainer
        trainer = Trainer(
            model=self.model,
            args=training_args,
            train_dataset=train_dataset,
            callbacks=callbacks,
            **trainer_extra_args,
        )

        # To do: do we provide feedback to the user?
        LOGGER.info("calling trainer.train")
        trainer.train()
        LOGGER.info("done with training")

        save_path = training_tmp_dir / tuned_model_name
        trainer.save_model(save_path)
        self.preprocessor.save_pretrained(save_path)
        return save_path


class TinyTimeMixerForecastingTuningHandler(TSFMForecastingTuningHandler):
    """Service handler for the tiny time mixer model"""

    def _get_config_kwargs(
        self,
        parameters: Optional[ForecastingParameters] = None,
        preprocessor: Optional[TimeSeriesPreprocessor] = None,
    ) -> Dict[str, Any]:
        config_kwargs = {
            "num_input_channels": preprocessor.num_input_channels,
            "prediction_filter_length": parameters.prediction_length,
            "exogenous_channel_indices": preprocessor.exogenous_channel_indices,
            "prediction_channel_indices": preprocessor.prediction_channel_indices,
        }
        return config_kwargs



# Standard

import tempfile
from pathlib import Path

import pytest
from tsfmfinetuning.finetuning import FinetuningRuntime
from tsfmfinetuning.ftpayloads import TinyTimeMixerForecastingTuneInput


PAYLOADS = [
    {
        "data": "file://./path/to/pandas.csv/or/pandas.feather/or/pandas.csv.gz",
        "model_id": "mytest-tsfm/ttm-r1",
        "schema": {
            "timestamp_column": "date",
            "target_columns": ["HUFL", "HULL", "MUFL", "MULL", "LUFL", "LULL", "OT"],
        },
        "parameters": {
            "tune_prefix": "fine_tuned/",
            "trainer_args": {"num_train_epochs": 1, "per_device_train_batch_size": 256},
            "fewshot_fraction": 0.05,
        },
    }
]

file_data_uris = [
    "file://./data/ETTh1.feather",
    "file://./data/ETTh1.csv",
    "file://./data/ETTh1.csv.gz",
    "file://./data/multipart",
]


@pytest.mark.parametrize("uri", file_data_uris)
@pytest.mark.parametrize("payload", PAYLOADS)
def test_fine_tune_forecasting_with_local_io(uri, payload):
    payload["data"] = uri
    input: TinyTimeMixerForecastingTuneInput = TinyTimeMixerForecastingTuneInput(**payload)

    ftr: FinetuningRuntime = FinetuningRuntime()

    with tempfile.TemporaryDirectory() as tmp_dir:
        tuned_model_name = "pytest_tuned_model"
        response = ftr.finetuning(input=input, tuned_model_name=tuned_model_name, output_dir=Path(tmp_dir))
        assert response.exists()
        assert (response / "config.json").exists() and (response / "config.json").stat().st_size > 0



import argparse
import tempfile
from pathlib import Path


def argparser():
    parser = argparse.ArgumentParser()

    # ################Input data
    parser.add_argument(
        "--payload",
        "-p",
        type=Path,
        help="A json file containing the service request payload.",
        required="true",
    )
    parser.add_argument(
        "--target_dir",
        "-d",
        type=Path,
        help="A target directory where the fine-tuned model should be written.",
        default=tempfile.gettempdir(),
    )
    parser.add_argument(
        "--model_name",
        "-n",
        type=str,
        help="The name that should be used for the finetuned model.",
        required=True,
    )
    # #################Model archicture
    arch_choices = ["ttm"]
    model_arch_type = parser.add_mutually_exclusive_group(
        required=True,
    )
    model_arch_type.add_argument(
        "--model_arch",
        "-a",
        choices=arch_choices,
        help=f"""The base model architecture. The following are currently supported: {",".join(arch_choices)}""",
    )

    # ####################Task type
    task_choices = ["forecasting"]
    model_task_choices = parser.add_mutually_exclusive_group(required=True)
    model_task_choices.add_argument(
        "--task_type",
        "-t",
        choices=task_choices,
        help=f"""The base model architecture. The following are currently supported: {",".join(task_choices)}""",
    )
    return parser
